from pathlib import Path

from dash import callback_context as ctx
from dash.exceptions import PreventUpdate


def toggle_modal_helper(
    n_clicks_close: int,
    n_clicks_open: int,
    is_open: bool,
    dont_show_again: bool,
    dsa_filepath: str,
    open_btn_id: str,
):
    if not ctx.triggered:
        raise PreventUpdate
    else:
        button_id = ctx.triggered[0]["prop_id"].split(".")[0]

    if button_id == open_btn_id:
        class_name = "hidden"
    else:
        class_name = None
    if dont_show_again:
        Path(dsa_filepath).touch()
    if n_clicks_close or n_clicks_open:
        return not is_open, class_name
    return is_open, class_name



import pandas as pd


def env_actions(episode, which="hazards", kind="ts", aggr=True):
    if kind not in ("ts", "nb", "dur"):
        raise ValueError(
            "kind argument can only be either ts, nb or dur.\n" f"{kind} passed"
        )
    if which not in ("hazards", "maintenances"):
        raise ValueError(
            "which argument can only be either hazards or "
            f"maintenances. {which} passed"
        )
    env_acts = getattr(episode, which)
    env_acts = env_acts.fillna(0)
    env_acts = pd.pivot_table(
        env_acts, index="timestamp", columns=["line_name"], values="value"
    )

    if kind == "dur":
        env_acts = env_acts.sum()
    if kind == "nb":
        equipments = env_acts.columns
        for col in env_acts.columns:
            env_acts["temp"] = env_acts[col].shift(1).fillna(0)
            env_acts[col] = env_acts[[col, "temp"]].apply(
                lambda row: 1 if row[0] == 1 and row[1] == 0 else 0, axis=1
            )
        env_acts = env_acts[equipments].sum()
        if aggr:
            env_acts = env_acts.sum()
    if kind == "ts" and aggr:
        env_acts = env_acts.sum(axis=1).to_frame(name=which)
    return env_acts



import os
import pathlib
import subprocess
import unittest


class TestMakeCache(unittest.TestCase):
    def setUp(self):
        self.agent_path = os.path.join(
            pathlib.Path(__file__).parent.absolute(),
            'data', 'agents')

    def test_make_cache(self):
        cmd = [
            'grid2viz',
            '--agents_path', self.agent_path,
            '--cache', 'True'
        ]
        rv = subprocess.run(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        self.assertEqual(rv.returncode, 0)



from functools import partial
import os
from pathlib import Path
import time

import dash_bootstrap_components as dbc
import dash_core_components as dcc
import dash_html_components as html
import plotly.graph_objects as go
from dash import callback_context
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
from pathos.multiprocessing import ProcessPool

from grid2viz.src.kpi import EpisodeTrace
from grid2viz.src.manager import (
    scenarios,
    best_agents,
    meta_json,
    make_episode_without_decorate,
    make_episode,
    n_cores,
    retrieve_episode_from_disk,
    save_in_ram_cache,
    cache_dir,
    grid2viz_home_directory,
)
from grid2viz.src.utils.callbacks_helpers import toggle_modal_helper
from grid2viz.src.utils.constants import DONT_SHOW_FILENAME


def register_callbacks_episodes(app):
    @app.callback(Output("cards_container", "children"), Input("url", "pathname"))
    def load_scenario_cards(url):
        """
        Create and display html cards with scenario's kpi for
        the 15 first scenarios using cache file.
        """
        cards_list = []
        cards_count = 0
        episode_graph_layout = {
            "autosize": True,
            "showlegend": False,
            "xaxis": {"showticklabels": False},
            "yaxis": {"showticklabels": False},
            "margin": {"l": 0, "r": 0, "t": 0, "b": 0},
        }

        url_split = url.split("/")
        url_split = url_split[len(url_split) - 1]

        is_episode_page = url_split == "/" or url_split == "" or url_split == "episodes"
        start_time = time.time()
        if cards_count < 15 and is_episode_page:
            sorted_scenarios = list(sorted(scenarios))
            if not os.path.exists(cache_dir):
                print(
                    "Starting Multiprocessing for reading the best agent of each scenario"
                )
                pool = ProcessPool(n_cores)
                best_agents_data = list(
                    pool.imap(
                        make_episode_without_decorate,
                        [
                            best_agents[scenario]["agent"]
                            for scenario in sorted_scenarios
                        ],
                        sorted_scenarios,
                    )
                )
                pool.close()
                print("Multiprocessing done")
                for i, scenario in enumerate(sorted_scenarios):
                    best_agent_episode = best_agents_data[i]
                    episode_data = retrieve_episode_from_disk(
                        best_agent_episode.episode_name, best_agent_episode.agent
                    )
                    best_agent_episode.decorate(episode_data)
                    save_in_ram_cache(
                        best_agent_episode.episode_name,
                        best_agent_episode.agent,
                        best_agent_episode,
                    )

            for i, scenario in enumerate(sorted_scenarios):
                best_agent_episode = make_episode(
                    best_agents[scenario]["agent"], scenario
                )
                prod_share = EpisodeTrace.get_prod_share_trace(best_agent_episode)
                consumption = best_agent_episode.profile_traces
                cards_list.append(
                    dbc.Col(
                        id=f"card_{scenario}",
                        lg=4,
                        width=12,
                        children=[
                            dbc.Card(
                                className="mb-3",
                                children=[
                                    dbc.CardBody(
                                        [
                                            html.H5(
                                                className="card-title",
                                                children="Scenario {0}".format(
                                                    scenario
                                                ),
                                            ),
                                            dbc.Row(
                                                children=[
                                                    dbc.Col(
                                                        className="mb-4",
                                                        children=[
                                                            html.P(
                                                                className="border-bottom h3 mb-0 text-right",
                                                                children=best_agents[
                                                                    scenario
                                                                ]["out_of"],
                                                            ),
                                                            html.P(
                                                                className="text-muted",
                                                                children="Agents on Scenario",
                                                            ),
                                                        ],
                                                    ),
                                                    dbc.Col(
                                                        className="mb-4",
                                                        children=[
                                                            html.P(
                                                                className="border-bottom h3 mb-0 text-right",
                                                                children="{}/{}".format(
                                                                    best_agents[
                                                                        scenario
                                                                    ]["value"],
                                                                    meta_json[scenario][
                                                                        "chronics_max_timestep"
                                                                    ],
                                                                ),
                                                            ),
                                                            html.P(
                                                                className="text-muted",
                                                                children="Agent's Survival",
                                                            ),
                                                        ],
                                                    ),
                                                    dbc.Col(
                                                        className="mb-4",
                                                        children=[
                                                            html.P(
                                                                className="border-bottom h3 mb-0 text-right",
                                                                children=f'{round(best_agents[scenario]["cum_reward"]):,}',
                                                            ),
                                                            html.P(
                                                                className="text-muted",
                                                                children="Cumulative Reward",
                                                            ),
                                                        ],
                                                    ),
                                                    dbc.Col(
                                                        className="mb-4",
                                                        children=[
                                                            html.P(
                                                                className="border-bottom h3 mb-0 text-right",
                                                                children="{} min".format(
                                                                    round(
                                                                        best_agent_episode.total_maintenance_duration
                                                                    )
                                                                ),
                                                            ),
                                                            html.P(
                                                                className="text-muted",
                                                                children="Total Maintenance Duration",
                                                            ),
                                                        ],
                                                    ),
                                                ]
                                            ),
                                            dbc.Row(
                                                className="align-items-center",
                                                children=[
                                                    dbc.Col(
                                                        lg=4,
                                                        width=12,
                                                        children=[
                                                            html.H5(
                                                                "Production Share",
                                                                className="text-center",
                                                            ),
                                                            dcc.Graph(
                                                                style={
                                                                    "height": "150px"
                                                                },
                                                                figure=go.Figure(
                                                                    layout=episode_graph_layout,
                                                                    data=prod_share,
                                                                ),
                                                            ),
                                                        ],
                                                    ),
                                                    dbc.Col(
                                                        lg=8,
                                                        width=12,
                                                        children=[
                                                            html.H5(
                                                                "Consumption Profile",
                                                                className="text-center",
                                                            ),
                                                            dcc.Graph(
                                                                style={
                                                                    "height": "150px"
                                                                },
                                                                figure=go.Figure(
                                                                    layout=episode_graph_layout,
                                                                    data=consumption,
                                                                ),
                                                            ),
                                                        ],
                                                    ),
                                                ],
                                            ),
                                        ]
                                    ),
                                    dbc.CardFooter(
                                        dbc.Button(
                                            "Open",
                                            id=scenario,
                                            key=scenario,
                                            className="btn-block",
                                            style={"background-color": "#2196F3"},
                                        )
                                    ),
                                ],
                            )
                        ],
                    )
                )
                cards_count += 1
        print(
            "Initial loading time for the best agent of all scenarios = {:.1f} seconds".format(
                time.time() - start_time
            )
        )
        return cards_list

    @app.callback(
        [Output("scenario", "data"), Output("url", "pathname")],
        [Input(scenario, "n_clicks") for scenario in scenarios],
        [State(scenario, "key") for scenario in scenarios],
    )
    def open_scenario(*input_state):
        """
        Open scenario into the overview layout when button
        corresponding button is clicked.

        Use callback context to get triggered input then parse it to get triggered input id
        then get the state key value from context with is the dict key (input_id + '.key').

        .. note:: you may need to see https://dash.plot.ly/faqs to get how I determine which Input has changed
        """

        ctx = callback_context

        # No clicks
        if not ctx.triggered:
            raise PreventUpdate
        # No clicks again
        # https://github.com/plotly/dash/issues/684
        if ctx.triggered[0]["value"] is None:
            raise PreventUpdate

        input_id = ctx.triggered[0]["prop_id"].split(".")[0]
        input_key = ctx.states[input_id + ".key"]
        scenario = input_key

        return scenario, "/overview"

    @app.callback(
        [
            Output("modal_episodes", "is_open"),
            Output("dont_show_again_div_episodes", "className"),
        ],
        [Input("close_episodes", "n_clicks"), Input("page_help", "n_clicks")],
        [
            State("modal_episodes", "is_open"),
            State("dont_show_again_episodes", "checked"),
        ],
    )
    def toggle_modal(close_n_clicks, open_n_clicks, is_open, dont_show_again):
        dsa_filepath = Path(grid2viz_home_directory) / DONT_SHOW_FILENAME("episodes")
        return toggle_modal_helper(
            close_n_clicks,
            open_n_clicks,
            is_open,
            dont_show_again,
            dsa_filepath,
            "page_help",
        )

    @app.callback(
        Output("collapse", "is_open"),
        [Input("collapse-button", "n_clicks")],
        [State("collapse", "is_open")],
    )
    def toggle_collapse(n, is_open):
        if n:
            return not is_open
        return is_open

    @app.callback(Output("modal_image_episodes", "src"), [Input("url", "pathname")])
    def show_image(pathname):
        return app.get_asset_url("screenshots/scenario_selection.png")

    @app.callback(
        Output("scenario_filter_div", "className"), [Input("collapse", "is_open")]
    )
    def toggle_scenario_filter(is_open):
        if is_open:
            return ""
        else:
            return "hidden"

    def filter_scenarios(selected_scenarios, scenario):
        if selected_scenarios is None:
            raise PreventUpdate
        if scenario in selected_scenarios:
            return ""
        else:
            return "hidden"

    for scenario in scenarios:
        app.callback(
            Output(f"card_{scenario}", "className"),
            [Input("scenarios_filter", "value")],
        )(partial(filter_scenarios, scenario=scenario))



import datetime as dt
from pathlib import Path

import plotly.graph_objects as go
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

from grid2viz.src.manager import grid2viz_home_directory
from grid2viz.src.manager import make_episode, make_network_agent_study
from grid2viz.src.utils import common_graph
from grid2viz.src.utils.callbacks_helpers import toggle_modal_helper
from grid2viz.src.utils.constants import DONT_SHOW_FILENAME
from grid2viz.src.utils.graph_utils import relayout_callback, get_axis_relayout


def register_callbacks_micro(app):
    @app.callback(
        [
            Output("slider", "min"),
            Output("slider", "max"),
            Output("slider", "value"),
            Output("slider", "marks"),
        ],
        [Input("window", "data")],
        [
            State("slider", "value"),
            State("user_timestamps", "value"),
            State("agent_study", "data"),
            State("scenario", "data"),
        ],
    )
    def update_slider(window, value, selected_timestamp, study_agent, scenario):
        if window is None:
            raise PreventUpdate
        new_episode = make_episode(study_agent, scenario)

        min_ = new_episode.timestamps.index(
            dt.datetime.strptime(window[0], "%Y-%m-%dT%H:%M:%S")
        )
        max_ = new_episode.timestamps.index(
            dt.datetime.strptime(window[1], "%Y-%m-%dT%H:%M:%S")
        )
        if value not in range(min_, max_):
            value = new_episode.timestamps.index(
                dt.datetime.strptime(selected_timestamp, "%Y-%m-%d %H:%M")
            )

        marks = dict(enumerate(map(lambda x: x.time(), new_episode.timestamps)))

        return min_, max_, value, marks

    @app.callback(
        Output("relayoutStoreMicro", "data"),
        [
            Input("env_charts_ts", "relayoutData"),
            Input("usage_rate_ts", "relayoutData"),
            Input("overflow_ts", "relayoutData"),
            Input("rewards_ts", "relayoutData"),
            Input("cumulated_rewards_ts", "relayoutData"),
            Input("actions_ts", "relayoutData"),
            Input("voltage_flow_graph", "relayoutData"),
        ],
        [State("relayoutStoreMicro", "data")],
    )
    def relayout_store_overview(*args):
        return relayout_callback(*args)

    # indicator line
    @app.callback(
        [
            Output("rewards_ts", "figure"),
            Output("cumulated_rewards_ts", "figure"),
            Output("actions_ts", "figure"),
        ],
        [
            Input("relayoutStoreMicro", "data"),
            Input("window", "data"),
            State("agent_study", "data"),
        ],
        [
            State("rewards_ts", "figure"),
            State("cumulated_rewards_ts", "figure"),
            State("actions_ts", "figure"),
            State("agent_ref", "data"),
            State("scenario", "data"),
        ],
    )
    def load_ts(
        relayout_data_store,
        window,
        study_agent,
        rew_figure,
        cumrew_figure,
        actions_figure,
        agent_ref,
        scenario,
    ):

        figures = [rew_figure, cumrew_figure, actions_figure]

        condition = (
            relayout_data_store is not None and relayout_data_store["relayout_data"]
        )
        if condition:
            relayout_data = relayout_data_store["relayout_data"]
            relayouted = False
            for figure in figures:
                axis_layout = get_axis_relayout(figure, relayout_data)
                if axis_layout is not None:
                    figure["layout"].update(axis_layout)
                    relayouted = True
            if relayouted:
                return figures

        rew_figure, cumrew_figure = common_graph.make_rewards_ts(
            study_agent, agent_ref, scenario, rew_figure, cumrew_figure
        )

        actions_figure = common_graph.make_action_ts(
            study_agent, agent_ref, scenario, actions_figure["layout"]
        )

        figures = [rew_figure, cumrew_figure, actions_figure]

        if window is not None:
            start_datetime = dt.datetime.strptime(window[0], "%Y-%m-%dT%H:%M:%S")
            end_datetime = dt.datetime.strptime(window[-1], "%Y-%m-%dT%H:%M:%S")
            for figure in figures:
                figure["layout"].update(
                    xaxis=dict(range=[start_datetime, end_datetime], autorange=False)
                )

        return figures

    # flux line callback
    @app.callback(
        [Output("line_side_choices", "options"), Output("line_side_choices", "value")],
        [Input("voltage_flow_choice", "value"), Input("flow_radio", "value")],
        [State("agent_study", "data"), State("scenario", "data")],
    )
    def load_voltage_flow_line_choice(category, flow_choice, study_agent, scenario):
        option = []
        new_episode = make_episode(study_agent, scenario)

        if category == "voltage":
            for name in new_episode.line_names:
                option.append({"label": "ex_" + name, "value": "ex_" + name})
                option.append({"label": "or_" + name, "value": "or_" + name})
        if category == "flow":
            for name in new_episode.line_names:
                if flow_choice == "active_flow":
                    option.append(
                        {"label": "ex_active_" + name, "value": "ex_active_" + name}
                    )
                    option.append(
                        {"label": "or_active_" + name, "value": "or_active_" + name}
                    )
                if flow_choice == "current_flow":
                    option.append(
                        {"label": "ex_current_" + name, "value": "ex_current_" + name}
                    )
                    option.append(
                        {"label": "or_current_" + name, "value": "or_current_" + name}
                    )

                if flow_choice == "flow_usage_rate":
                    option.append(
                        {"label": "usage_rate_" + name, "value": "usage_rate_" + name}
                    )
        if category == "redispatch":
            option = [
                {"label": gen_name, "value": gen_name}
                for gen_name in new_episode.prod_names
            ]

        return option, [option[0]["value"]]

    @app.callback(
        Output("voltage_flow_graph", "figure"),
        [
            Input("line_side_choices", "value"),
            Input("voltage_flow_choice", "value"),
            Input("relayoutStoreMicro", "data"),
            Input("window", "data"),
        ],
        [
            State("voltage_flow_graph", "figure"),
            State("agent_study", "data"),
            State("scenario", "data"),
        ],
    )
    def load_flow_voltage_graph(
        selected_objects,
        choice,
        relayout_data_store,
        window,
        figure,
        study_agent,
        scenario,
    ):
        if relayout_data_store is not None and relayout_data_store["relayout_data"]:
            relayout_data = relayout_data_store["relayout_data"]
            layout = figure["layout"]
            new_axis_layout = get_axis_relayout(figure, relayout_data)
            if new_axis_layout is not None:
                layout.update(new_axis_layout)
                return figure
        new_episode = make_episode(study_agent, scenario)
        if selected_objects is not None:
            if choice == "voltage":
                figure["data"] = load_voltage_for_lines(selected_objects, new_episode)
            if "flow" in choice:
                figure["data"] = load_flows_for_lines(selected_objects, new_episode)
            if "redispatch" in choice:
                figure["data"] = load_redispatch(selected_objects, new_episode)

        if window is not None:
            figure["layout"].update(xaxis=dict(range=window, autorange=False))

        return figure

    @app.callback(
        Output("flow_radio", "style"),
        [Input("voltage_flow_choice", "value")],
    )
    def load_flow_graph(choice):
        if choice == "flow":
            return {"display": "block"}
        else:
            return {"display": "none"}

    def load_voltage_for_lines(lines, new_episode):
        voltage = new_episode.flow_and_voltage_line
        traces = []

        for value in lines:
            # the first 2 characters are the side of line ('ex' or 'or')
            line_side = str(value)[:2]
            line_name = str(value)
            if line_side == "ex":
                traces.append(
                    go.Scatter(
                        x=new_episode.timestamps,
                        # remove the first 3 char to get the line name and round to 3 dec
                        y=voltage["ex"]["voltage"][line_name[3:]],
                        name=line_name,
                    )
                )
            if line_side == "or":
                traces.append(
                    go.Scatter(
                        x=new_episode.timestamps,
                        y=voltage["or"]["voltage"][line_name[3:]],
                        name=line_name,
                    )
                )
        return traces

    def load_redispatch(generators, new_episode):
        actual_dispatch = new_episode.actual_redispatch
        target_dispatch = new_episode.target_redispatch
        traces = []

        x = new_episode.timestamps

        for gen in generators:
            traces.append(
                go.Scatter(x=x, y=actual_dispatch[gen], name=f"{gen} actual dispatch")
            )
            traces.append(
                go.Scatter(x=x, y=target_dispatch[gen], name=f"{gen} target dispatch")
            )

        return traces

    def load_flows_for_lines(lines, new_episode):
        flow = new_episode.flow_and_voltage_line
        traces = []

        x = new_episode.timestamps

        for value in lines:
            line_side = str(value)[
                :2
            ]  # the first 2 characters are the side of line ('ex' or 'or')
            flow_type = str(value)[3:].split("_", 1)[
                0
            ]  # the type is the 1st part of the string: 'type_name'
            line_name = str(value)[3:].split("_", 1)[
                1
            ]  # the name is the 2nd part of the string: 'type_name'
            if line_side == "ex":
                traces.append(
                    go.Scatter(x=x, y=flow["ex"][flow_type][line_name], name=value)
                )
            elif line_side == "or":
                traces.append(
                    go.Scatter(x=x, y=flow["or"][flow_type][line_name], name=value)
                )
            else:  # this concern usage rate
                name = value.split("_", 2)[2]  # get the powerline name
                index_powerline = list(new_episode.line_names).index(name)
                usage_rate_powerline = new_episode.rho.loc[
                    new_episode.rho["equipment"] == index_powerline
                ]["value"]

                traces.append(go.Scatter(x=x, y=usage_rate_powerline, name=name))

        return traces

    # context line callback
    @app.callback(
        [Output("asset_selector", "options"), Output("asset_selector", "value")],
        [Input("environment_choices_buttons", "value")],
        [State("agent_study", "data"), State("scenario", "data")],
    )
    def update_ts_graph_avail_assets(kind, study_agent, scenario):
        new_episode = make_episode(study_agent, scenario)
        return common_graph.ts_graph_avail_assets(kind, new_episode)

    @app.callback(
        Output("env_charts_ts", "figure"),
        [
            Input("asset_selector", "value"),
            Input("relayoutStoreMicro", "data"),
            Input("window", "data"),
        ],
        [
            State("env_charts_ts", "figure"),
            State("environment_choices_buttons", "value"),
            State("scenario", "data"),
            State("agent_study", "data"),
        ],
    )
    def load_context_data(
        equipments, relayout_data_store, window, figure, kind, scenario, agent_study
    ):
        if relayout_data_store is not None and relayout_data_store["relayout_data"]:
            relayout_data = relayout_data_store["relayout_data"]
            layout = figure["layout"]
            new_axis_layout = get_axis_relayout(figure, relayout_data)
            if new_axis_layout is not None:
                layout.update(new_axis_layout)
                return figure

        if kind is None:
            return figure
        if isinstance(equipments, str):
            equipments = [equipments]  # to make pd.series.isin() work
        episode = make_episode(agent_study, scenario)
        figure["data"] = common_graph.environment_ts_data(kind, episode, equipments)

        if window is not None:
            figure["layout"].update(xaxis=dict(range=window, autorange=False))

        return figure

    @app.callback(
        [Output("overflow_ts", "figure"), Output("usage_rate_ts", "figure")],
        [Input("relayoutStoreMicro", "data"), Input("window", "data")],
        [
            State("overflow_ts", "figure"),
            State("usage_rate_ts", "figure"),
            State("agent_study", "data"),
            State("scenario", "data"),
        ],
    )
    def update_agent_ref_graph(
        relayout_data_store,
        window,
        figure_overflow,
        figure_usage,
        study_agent,
        scenario,
    ):
        if relayout_data_store is not None and relayout_data_store["relayout_data"]:
            relayout_data = relayout_data_store["relayout_data"]
            layout_usage = figure_usage["layout"]
            new_axis_layout = get_axis_relayout(figure_usage, relayout_data)
            if new_axis_layout is not None:
                layout_usage.update(new_axis_layout)
                figure_overflow["layout"].update(new_axis_layout)
                return figure_overflow, figure_usage

        if window is not None:
            figure_overflow["layout"].update(xaxis=dict(range=window, autorange=False))
            figure_usage["layout"].update(xaxis=dict(range=window, autorange=False))

        return common_graph.agent_overflow_usage_rate_trace(
            make_episode(study_agent, scenario), figure_overflow, figure_usage
        )

    @app.callback(
        Output("timeseries_table_micro", "data"), [Input("timeseries_table", "data")]
    )
    def sync_timeseries_table(data):
        return data

    @app.callback(
        [
            Output("interactive_graph", "figure"),
            Output("tooltip_table_micro", "children"),
        ],
        [Input("slider", "value")],
        [State("agent_study", "data"), State("scenario", "data")],
    )
    def update_interactive_graph(slider_value, study_agent, scenario):
        new_episode = make_episode(study_agent, scenario)
        act = new_episode.actions[slider_value]
        if any(act.get_types()):
            act_as_str = str(act)
        else:
            act_as_str = "NO ACTION"
        return make_network_agent_study(new_episode, timestep=slider_value), act_as_str

    @app.callback(
        [
            Output("modal_micro", "is_open"),
            Output("dont_show_again_div_micro", "className"),
        ],
        [Input("close_micro", "n_clicks"), Input("page_help", "n_clicks")],
        [State("modal_micro", "is_open"), State("dont_show_again_micro", "checked")],
    )
    def toggle_modal(close_n_clicks, open_n_clicks, is_open, dont_show_again):
        dsa_filepath = Path(grid2viz_home_directory) / DONT_SHOW_FILENAME("micro")
        return toggle_modal_helper(
            close_n_clicks,
            open_n_clicks,
            is_open,
            dont_show_again,
            dsa_filepath,
            "page_help",
        )

    @app.callback(Output("modal_image_micro", "src"), [Input("url", "pathname")])
    def show_image(pathname):
        return app.get_asset_url("screenshots/agent_study.png")



import pandas as pd
import plotly.graph_objects as go

from .observation_model import quantile10, quantile25, quantile75, quantile90


def consumption_profiles(episode, freq="30T"):

    # filter intercos
    idx_no_interco = [
        i
        for i in range(len(episode.load.equipment_name))
        if "interco" not in episode.load.equipment_name[i]
    ]

    load = (
        pd.pivot_table(
            episode.load.loc[idx_no_interco],
            index="timestamp",
            columns=["equipment_name"],
            values="value",
        )
        .sum(axis=1)
        .resample(freq)
        .mean()
        .to_frame(name="load")
        .reset_index()
    )

    if freq == "H":
        load["timestamp"] = load.timestamp.dt.hour
    elif "T" in freq:
        load["timestamp"] = load.timestamp.dt.time.map(lambda x: x.strftime("%H:%M"))
    else:
        raise ValueError(
            'Only hourly ("H") or minutely ("xT") freqs are '
            f"supported. Passed: {freq}"
        )

    daily_load_distrib = (
        load.groupby(["timestamp"])
        .aggregate(["median", quantile10, quantile25, quantile75, quantile90, "max"])[
            ["load"]
        ]
        .reset_index()
    )

    return daily_load_distrib


def profiles_traces(episode, freq="30T"):
    episode_data = episode
    df = consumption_profiles(episode_data, freq)
    line = {"shape": "spline", "width": 0, "smoothing": 1}
    trace = [
        go.Scatter(
            x=df["timestamp"], y=df["load"]["quantile10"], name="quantile 10", line=line
        ),
        go.Scatter(
            x=df["timestamp"],
            y=df["load"]["quantile25"],
            name="quantile 25",
            fill="tonexty",
            fillcolor="rgba(159, 197, 232, 0.63)",
            line=line,
        ),
        go.Scatter(
            x=df["timestamp"],
            y=df["load"]["median"],
            name="median",
            fill="tonexty",
            fillcolor="rgba(31, 119, 180, 0.5)",
            line={"color": "rgb(31, 119, 180)", "shape": "spline", "smoothing": 1},
        ),
        go.Scatter(
            x=df["timestamp"],
            y=df["load"]["quantile75"],
            name="quantile 75",
            fill="tonexty",
            fillcolor="rgba(31, 119, 180, 0.5)",
            line=line,
        ),
        go.Scatter(
            x=df["timestamp"],
            y=df["load"]["quantile90"],
            name="quantile 90",
            fill="tonexty",
            fillcolor="rgba(159, 197, 232, 0.63)",
            line=line,
        ),
        go.Scatter(
            x=df["timestamp"],
            y=df["load"]["max"],
            name="Max",
            line={"shape": "spline", "smoothing": 1, "color": "rgba(255,0,0,0.5)"},
        ),
    ]
    return trace



import datetime as dt
import time

import numpy as np
import pandas as pd
from grid2op.Episode import EpisodeData
from tqdm import tqdm

from . import EpisodeTrace, maintenances, consumption_profiles
from .env_actions import env_actions

# TODO: configure the reward key you want to visualize in agent overview.
# Either as an argument or a dropdown list in the app from which we can choose.
# The reward dataframe should get bigger with all keys available anyway
other_reward_key = "grid_operation_cost"


class ActionImpacts:
    def __init__(
        self,
        action_line,
        action_subs,
        action_redisp,
        redisp_impact,
        line_name,
        sub_name,
        gen_name,
        action_id,
    ):
        self.action_line = action_line
        self.action_subs = action_subs
        self.action_redisp = action_redisp
        self.redisp_impact = redisp_impact
        self.line_name = line_name
        self.sub_name = sub_name
        self.gen_name = gen_name
        self.action_id = action_id


class EpisodeAnalytics:
    def __init__(self, episode_data, episode_name, agent):
        self.episode_name = episode_name
        self.agent = agent

        self.timesteps = list(range(len(episode_data.actions)))
        print(
            f"Computing DataFrames for scenario {self.episode_name} and agent {self.agent}"
        )
        beg = time.time()
        print("Environment")
        (
            self.load,
            self.production,
            self.rho,
            self.action_data_table,
            self.computed_reward,
            self.flow_and_voltage_line,
            self.target_redispatch,
            self.actual_redispatch,
            self.attacks_data_table,
        ) = self._make_df_from_data(episode_data)
        print("Hazards-Maintenances")
        self.hazards, self.maintenances = self._env_actions_as_df(episode_data)
        print("Computing computation intensive indicators...")
        self.total_overflow_trace = EpisodeTrace.get_total_overflow_trace(
            self, episode_data
        )
        self.usage_rate_trace = EpisodeTrace.get_usage_rate_trace(self)
        self.reward_trace = EpisodeTrace.get_df_rewards_trace(self)
        self.total_overflow_ts = EpisodeTrace.get_total_overflow_ts(self, episode_data)
        self.profile_traces = consumption_profiles.profiles_traces(self)
        self.total_maintenance_duration = maintenances.total_duration_maintenance(self)
        self.nb_hazards = env_actions(self, which="hazards", kind="nb", aggr=True)
        self.nb_maintenances = env_actions(
            self, which="maintenances", kind="nb", aggr=True
        )

        end = time.time()
        print(f"end computing df: {end - beg}")

    @staticmethod
    def timestamp(obs):
        return dt.datetime(
            obs.year, obs.month, obs.day, obs.hour_of_day, obs.minute_of_hour
        )

    # @jit(forceobj=True)
    def _make_df_from_data(self, episode_data):
        """
        Convert all episode's data into comprehensible dataframes usable by
        the application.

        The generated dataframes are:
            - loads
            - production
            - rho
            - action data table
            - instant and cumulated rewards
            - flow and voltage by line
            - target and actual redispatch
            - attacks

        Returns
        -------
        res: :class:`tuple`
         generated dataframes
        """
        size = len(episode_data.actions)
        timesteps = list(range(size))
        load_size = size * len(episode_data.observations[0].load_p)
        prod_size = size * len(episode_data.observations[0].prod_p)
        n_rho = len(episode_data.observations[0].rho)
        rho_size = size * n_rho

        load_data = pd.DataFrame(index=range(load_size), columns=["timestamp", "value"])
        load_data.loc[:, "value"] = load_data.loc[:, "value"].astype(float)

        production = pd.DataFrame(index=range(prod_size), columns=["value"])

        rho = pd.DataFrame(index=range(rho_size), columns=["value"])

        cols_loop_action_data_table = [
            "action_line",
            "action_subs",
            "action_redisp",
            "redisp_impact",
            "line_name",
            "sub_name",
            "gen_name",
            "action_id",
            "distance",
            "lines_modified",
            "subs_modified",
            "gens_modified",
        ]
        action_data_table = pd.DataFrame(
            index=range(size),
            columns=[
                "timestep",
                "timestamp",
                "timestep_reward",
                "action_line",
                "action_subs",
                "action_redisp",
                "redisp_impact",
                "line_name",
                "sub_name",
                "gen_name",
                "action_id",
                "distance",
                "lines_modified",
                "subs_modified",
                "gens_modified",
            ],
        )

        computed_rewards = pd.DataFrame(
            index=range(size), columns=["timestep", "rewards", "cum_rewards"]
        )
        flow_voltage_cols = pd.MultiIndex.from_product(
            [
                ["or", "ex"],
                ["active", "reactive", "current", "voltage"],
                episode_data.line_names,
            ]
        )
        flow_voltage_line_table = pd.DataFrame(
            index=range(size), columns=flow_voltage_cols
        )

        target_redispatch = pd.DataFrame(
            index=range(size), columns=episode_data.prod_names
        )
        actual_redispatch = pd.DataFrame(
            index=range(size), columns=episode_data.prod_names
        )

        topo_vect = episode_data.observations[0].topo_vect
        if topo_vect.sum() != len(topo_vect):
            raise ValueError("Not all things are on bus 1")

        obs_0 = episode_data.observations[0]

        # True == connected, False == disconnect
        # So that len(line_statuses) - line_statuses.sum() is the distance for lines
        line_statuses = episode_data.observations[0].line_status

        # True == sub has something on bus 2, False == everything on bus 1
        # So that subs_on_bus2.sum() is the distance for subs
        subs_on_bus_2 = np.repeat(False, episode_data.observations[0].n_sub)

        # objs_on_bus_2 will store the id of objects connected to bus 2
        objs_on_bus_2 = {id: [] for id in range(episode_data.observations[0].n_sub)}

        # Distance from original topology is then :
        # len(line_statuses) - line_statuses.sum() + subs_on_bus_2.sum()

        gens_modified_ids = []
        actual_redispatch_previous_ts = obs_0.actual_dispatch

        list_actions = []
        for (time_step, (obs, act)) in tqdm(
            enumerate(zip(episode_data.observations[:-1], episode_data.actions)),
            total=size,
        ):
            time_stamp = self.timestamp(obs)
            (
                action_impacts,
                list_actions,
                lines_modified,
                subs_modified,
                gens_modified_names,
                gens_modified_ids,
            ) = self.compute_action_impacts(
                act, list_actions, obs, gens_modified_ids, actual_redispatch_previous_ts
            )

            actual_redispatch_previous_ts = obs.actual_dispatch

            # Building load DF
            begin = time_step * episode_data.n_loads
            end = (time_step + 1) * episode_data.n_loads - 1
            load_data.loc[begin:end, "value"] = obs.load_p.astype(float)
            load_data.loc[begin:end, "timestamp"] = time_stamp
            # Building prod DF
            begin = time_step * episode_data.n_prods
            end = (time_step + 1) * episode_data.n_prods - 1
            production.loc[begin:end, "value"] = obs.prod_p.astype(float)
            # Building RHO DF
            begin = time_step * n_rho
            end = (time_step + 1) * n_rho - 1
            rho.loc[begin:end, "value"] = obs.rho.astype(float)

            pos = time_step

            (
                distance,
                line_statuses,
                subs_on_bus_2,
                objs_on_bus_2,
            ) = self.get_distance_from_obs(
                act, line_statuses, subs_on_bus_2, objs_on_bus_2, obs_0
            )

            action_data_table.loc[pos, cols_loop_action_data_table] = [
                action_impacts.action_line,
                action_impacts.action_subs,
                action_impacts.action_redisp,
                action_impacts.redisp_impact,
                action_impacts.line_name,
                action_impacts.sub_name,
                action_impacts.gen_name,
                action_impacts.action_id,
                distance,
                lines_modified,
                subs_modified,
                gens_modified_names,
            ]

            flow_voltage_line_table.loc[time_step, :] = np.array(
                [
                    obs.p_ex,
                    obs.q_ex,
                    obs.a_ex,
                    obs.v_ex,
                    obs.p_or,
                    obs.q_or,
                    obs.a_or,
                    obs.v_or,
                ]
            ).flatten()

            target_redispatch.loc[time_step, :] = obs.target_dispatch
            actual_redispatch.loc[time_step, :] = obs.actual_dispatch

        load_data["timestep"] = np.repeat(timesteps, episode_data.n_loads)
        load_data["equipment_name"] = np.tile(episode_data.load_names, size).astype(str)
        load_data["equipement_id"] = np.tile(range(episode_data.n_loads), size)

        self.timestamps = sorted(load_data.timestamp.dropna().unique())
        self.timesteps = sorted(load_data.timestep.unique())

        production["timestep"] = np.repeat(timesteps, episode_data.n_prods)
        production["timestamp"] = np.repeat(self.timestamps, episode_data.n_prods)
        production.loc[:, "equipment_name"] = np.tile(episode_data.prod_names, size)
        production.loc[:, "equipement_id"] = np.tile(range(episode_data.n_prods), size)

        rho["time"] = np.repeat(timesteps, n_rho)
        rho["timestamp"] = np.repeat(self.timestamps, n_rho)
        rho["equipment"] = np.tile(range(n_rho), size)

        action_data_table["timestep"] = self.timesteps
        action_data_table["timestamp"] = self.timestamps
        action_data_table["timestep_reward"] = episode_data.rewards[:size]

        load_data["value"] = load_data["value"].astype(float)
        production["value"] = production["value"].astype(float)
        rho["value"] = rho["value"].astype(float)

        computed_rewards["timestep"] = self.timestamps
        computed_rewards["rewards"] = episode_data.rewards[:size]

        # TODO: we should give a choice to select different rewards among other rewards
        if episode_data.other_rewards:
            if other_reward_key:
                if other_reward_key in episode_data.other_rewards[0].keys():
                    computed_rewards["rewards"] = [
                        other_reward[other_reward_key]
                        for other_reward in episode_data.other_rewards
                    ]
                    computed_rewards["rewards"] = computed_rewards["rewards"][:size]
        computed_rewards["cum_rewards"] = computed_rewards["rewards"].cumsum(axis=0)

        attacks_data_table = pd.DataFrame(
            index=range(size), columns=["timestep", "timestamp", "attack", "id_lines"]
        )
        attacks_data_table["timestep"] = self.timesteps
        attacks_data_table["timestamp"] = self.timestamps
        for time_step, attack in enumerate(episode_data.attacks):
            (
                n_lines_modified,
                str_lines_modified,
                lines_modified,
            ) = self.get_lines_modifications(attack)
            n_subs_modified, *_ = self.get_subs_modifications(attack)
            is_attacked = n_lines_modified > 0 or n_subs_modified > 0
            attacks_data_table.loc[time_step, "attack"] = is_attacked
            if len(lines_modified) == 0:
                attacks_data_table.loc[time_step, "id_lines"] = ""
            else:
                attacks_data_table.loc[time_step, "id_lines"] = lines_modified[0]

        return (
            load_data,
            production,
            rho,
            action_data_table,
            computed_rewards,
            flow_voltage_line_table,
            target_redispatch,
            actual_redispatch,
            attacks_data_table,
        )

    @staticmethod
    def get_action_id(action, list_actions):
        if not action:
            return None, list_actions
        for idx, act_dict in enumerate(list_actions):
            if action == act_dict:
                return idx, list_actions
        # if we havnt found the vect...
        list_actions.append(action)
        return len(list_actions) - 1, list_actions

    def get_sub_name(self, act, obs):
        for sub in range(len(obs.sub_info)):
            effect = act.effect_on(substation_id=sub)
            if np.any(effect["change_bus"] is True):
                return self.name_sub[sub]
            if np.any(effect["set_bus"] == 1) or np.any(effect["set_bus"] == -1):
                return self.name_sub[sub]
        return None

    def get_distance_from_obs(
        self, act, line_statuses, subs_on_bus_2, objs_on_bus_2, obs
    ):

        impact_on_objs = act.impact_on_objects()

        # lines reconnetions/disconnections
        line_statuses[
            impact_on_objs["force_line"]["disconnections"]["powerlines"]
        ] = False
        line_statuses[
            impact_on_objs["force_line"]["reconnections"]["powerlines"]
        ] = True
        line_statuses[impact_on_objs["switch_line"]["powerlines"]] = np.invert(
            line_statuses[impact_on_objs["switch_line"]["powerlines"]]
        )

        topo_vect_dict = {
            "load": obs.load_pos_topo_vect,
            "generator": obs.gen_pos_topo_vect,
            "line (extremity)": obs.line_ex_pos_topo_vect,
            "line (origin)": obs.line_or_pos_topo_vect,
        }

        # Bus manipulation
        if impact_on_objs["topology"]["changed"]:
            for modif_type in ["bus_switch", "assigned_bus"]:

                for elem in impact_on_objs["topology"][modif_type]:
                    objs_on_bus_2 = self.update_objs_on_bus(
                        objs_on_bus_2, elem, topo_vect_dict, kind=modif_type
                    )

            for elem in impact_on_objs["topology"]["disconnect_bus"]:
                # Disconnected bus counts as one for the distance
                subs_on_bus_2[elem["substation"]] = True

        subs_on_bus_2 = [
            True if objs_on_2 else False for _, objs_on_2 in objs_on_bus_2.items()
        ]

        distance = len(line_statuses) - line_statuses.sum() + sum(subs_on_bus_2)
        return distance, line_statuses, subs_on_bus_2, objs_on_bus_2

    def update_objs_on_bus(self, objs_on_bus_2, elem, topo_vect_dict, kind):
        for object_type, pos_topo_vect in topo_vect_dict.items():
            if elem["object_type"] == object_type and elem["bus"]:
                if kind == "bus_switch":
                    objs_on_bus_2 = self.update_objs_on_bus_switch(
                        objs_on_bus_2, elem, pos_topo_vect
                    )
                else:
                    objs_on_bus_2 = self.update_objs_on_bus_assign(
                        objs_on_bus_2, elem, pos_topo_vect
                    )
                break
        return objs_on_bus_2

    @staticmethod
    def update_objs_on_bus_switch(objs_on_bus_2, elem, pos_topo_vect):
        if pos_topo_vect[elem["object_id"]] in objs_on_bus_2[elem["substation"]]:
            # elem was on bus 2, remove it from objs_on_bus_2
            objs_on_bus_2[elem["substation"]] = [
                x
                for x in objs_on_bus_2[elem["substation"]]
                if x != pos_topo_vect[elem["object_id"]]
            ]
        else:
            objs_on_bus_2[elem["substation"]].append(pos_topo_vect[elem["object_id"]])
        return objs_on_bus_2

    @staticmethod
    def update_objs_on_bus_assign(objs_on_bus_2, elem, pos_topo_vect):
        if (
            pos_topo_vect[elem["object_id"]] in objs_on_bus_2[elem["substation"]]
            and elem["bus"] == 1
        ):
            # elem was on bus 2, remove it from objs_on_bus_2
            objs_on_bus_2[elem["substation"]] = [
                x
                for x in objs_on_bus_2[elem["substation"]]
                if x != pos_topo_vect[elem["object_id"]]
            ]
        elif (
            pos_topo_vect[elem["object_id"]] not in objs_on_bus_2[elem["substation"]]
            and elem["bus"] == 2
        ):
            objs_on_bus_2[elem["substation"]].append(pos_topo_vect[elem["object_id"]])
        return objs_on_bus_2

    # @jit(forceobj=True)
    def _env_actions_as_df(self, episode_data):
        agent_length = len(
            episode_data.actions
        )  # int(episode_data.meta['nb_timestep_played'])
        hazards_size = agent_length * episode_data.n_lines
        cols = ["timestep", "timestamp", "line_id", "line_name", "value"]
        hazards = pd.DataFrame(index=range(hazards_size), columns=["value"], dtype=int)
        maintenances = hazards.copy()

        for (time_step, env_act) in tqdm(
            enumerate(episode_data.env_actions), total=len(episode_data.env_actions)
        ):
            if env_act is None:
                continue

            time_stamp = self.timestamp(episode_data.observations[time_step])

            begin = time_step * episode_data.n_lines
            end = (time_step + 1) * episode_data.n_lines - 1
            hazards.loc[begin:end, "value"] = env_act._hazards.astype(int)

            begin = time_step * episode_data.n_lines
            end = (time_step + 1) * episode_data.n_lines - 1
            maintenances.loc[begin:end, "value"] = env_act._maintenance.astype(int)

        hazards["timestep"] = np.repeat(range(agent_length), episode_data.n_lines)
        maintenances["timestep"] = hazards["timestep"]
        hazards["timestamp"] = np.repeat(self.timestamps, episode_data.n_lines)
        maintenances["timestamp"] = hazards["timestamp"]
        hazards["line_name"] = np.tile(episode_data.line_names, agent_length)
        maintenances["line_name"] = hazards["line_name"]
        hazards["line_id"] = np.tile(range(episode_data.n_lines), agent_length)
        maintenances["line_id"] = hazards["line_id"]

        return hazards, maintenances

    def get_prod_types(self):
        types = self.observation_space.gen_type
        ret = {}
        if types is None:
            return ret
        for (idx, name) in enumerate(self.prod_names):
            ret[name] = types[idx]
        return ret

    def decorate(self, episode_data):
        # Add EpisodeData attributes to EpisodeAnalytics
        for attribute in [
            elem
            for elem in dir(episode_data)
            if not (elem.startswith("__") or callable(getattr(episode_data, elem)))
        ]:
            setattr(self, attribute, getattr(episode_data, attribute))

    def compute_action_impacts(
        self,
        action,
        list_actions,
        observation,
        gens_modified_ids,
        actual_dispatch_previous_ts,
    ):

        (
            n_lines_modified,
            str_lines_modified,
            lines_modified,
        ) = self.get_lines_modifications(action)
        n_subs_modified, str_subs_modified, subs_modified = self.get_subs_modifications(
            action
        )

        (
            n_gens_modified,
            str_gens_modified,
            gens_modified_names,
            gens_modified_ids,
            redisp_volume,
        ) = self.get_gens_modifications(
            action, observation, gens_modified_ids, actual_dispatch_previous_ts
        )

        action_id, list_actions = self.get_action_id(action, list_actions)

        return (
            ActionImpacts(
                action_line=n_lines_modified,
                action_subs=n_subs_modified,
                action_redisp=n_gens_modified,
                redisp_impact=redisp_volume,
                line_name=str_lines_modified,
                sub_name=str_subs_modified,
                gen_name=str_gens_modified,
                action_id=action_id,
            ),
            list_actions,
            lines_modified,
            subs_modified,
            gens_modified_names,
            gens_modified_ids,
        )

    def get_lines_modifications(self, action):
        action_dict = action.as_dict()
        n_lines_modified = 0
        lines_reconnected = []
        lines_disconnected = []
        lines_switched = []
        str_lines_modified = ""
        if "set_line_status" in action_dict:
            n_lines_modified += (
                action_dict["set_line_status"]["nb_connected"]
                + action_dict["set_line_status"]["nb_disconnected"]
            )
            lines_reconnected = [
                *lines_reconnected,
                *[
                    action.name_line[int(line_id)]
                    for line_id in action_dict["set_line_status"]["connected_id"]
                ],
            ]
            if lines_reconnected:
                str_lines_modified += "Reconnect: " + ", ".join(lines_reconnected)
            lines_disconnected = [
                *lines_disconnected,
                *[
                    action.name_line[int(line_id)]
                    for line_id in action_dict["set_line_status"]["disconnected_id"]
                ],
            ]
            if lines_disconnected:
                if str_lines_modified:
                    str_lines_modified += " - "
                str_lines_modified += "Disconnect: " + ", ".join(lines_disconnected)
        if "change_line_status" in action_dict:
            n_lines_modified += action_dict["change_line_status"]["nb_changed"]
            lines_switched = [
                *lines_switched,
                *[
                    action.name_line[int(line_id)]
                    for line_id in action_dict["change_line_status"]["changed_id"]
                ],
            ]
            if lines_switched:
                if str_lines_modified:
                    str_lines_modified += " - "
                str_lines_modified += "Switch: " + ", ".join(lines_switched)

        lines_modified = [*lines_reconnected, *lines_disconnected, *lines_switched]

        return n_lines_modified, str_lines_modified, lines_modified

    def get_subs_modifications(self, action):
        action_dict = action.as_dict()
        n_subs_modified = 0
        subs_modified = []

        if "set_bus_vect" in action_dict:
            n_subs_modified += action_dict["set_bus_vect"]["nb_modif_subs"]
            subs_modified = [
                *subs_modified,
                *[
                    action.name_sub[int(sub_id)]
                    for sub_id in action_dict["set_bus_vect"]["modif_subs_id"]
                ],
            ]
        if "change_bus_vect" in action_dict:
            n_subs_modified += action_dict["change_bus_vect"]["nb_modif_subs"]
            subs_modified = [
                *subs_modified,
                *[
                    action.name_sub[int(sub_id)]
                    for sub_id in action_dict["change_bus_vect"]["modif_subs_id"]
                ],
            ]

        subs_modified_set = set(subs_modified)
        str_subs_modified = " - ".join(subs_modified_set)
        return n_subs_modified, str_subs_modified, subs_modified

    def get_gens_modifications(
        self,
        action,
        observation,
        gens_modified_previous_time_step,
        actual_dispatch_previous_ts,
    ):
        action_dict = action.as_dict()
        n_gens_modified = 0
        gens_modified_ids = []
        gens_modified_names = []
        if "redispatch" in action_dict:
            n_gens_modified = (action_dict["redispatch"] != 0).sum()
            gens_modified_ids = np.where(action_dict["redispatch"] != 0)[0]
            gens_modified_names = action.name_gen[gens_modified_ids]

        str_gens_modified = " - ".join(gens_modified_names)

        volume_redispatched = round(
            np.absolute(
                observation.actual_dispatch[gens_modified_previous_time_step]
                - actual_dispatch_previous_ts[gens_modified_previous_time_step]
            ).sum(),
            2,
        )

        return (
            n_gens_modified,
            str_gens_modified,
            gens_modified_names,
            gens_modified_ids,
            volume_redispatched,
        )

    def get_subs_and_lines_impacted(self, action):
        line_impact, sub_impact = action.get_topological_impact()
        sub_names = action.name_sub[sub_impact]
        line_names = action.name_line[line_impact]
        return sub_names, line_names

    def format_subs_and_lines_impacted(self, sub_names, line_names):
        return self.format_elements_impacted(sub_names), self.format_elements_impacted(
            line_names
        )

    def format_elements_impacted(self, elements):
        if not len(elements):
            elements_formatted = None
        else:
            elements_formatted = " - ".join(elements)
        return elements_formatted


class Test:
    def __init__(self):
        self.foo = 2
        self.bar = 3


if __name__ == "__main__":
    test = Test()
    path_agent = "nodisc_badagent"
    episode = EpisodeData.from_disk(
        "D:/Projects/RTE - Grid2Viz/20200127_data_scripts/20200127_agents_log/"
        + path_agent,
        "3_with_hazards",
    )
    print(dir(EpisodeAnalytics(episode)))



import os
import subprocess
import signal
import sys
import time
import pkg_resources
import argparse

CONFIG_FILE_CONTENT = """
# This file have been automatically generated, please do not modify it. 
# You can remove it once done with the application.
[DEFAULT]
agents_dir={agents_dir}
env_dir={env_dir}
n_cores={n_cores}
# This file will be re generated to each call of "python -m grid2viz.main"
"""

ARG_AGENTS_PATH_DESC = 'The path where the log of the Agents experiences are stored.' \
                       ' (default to None to study the example agents provided with the package)'
ARG_ENV_PATH_DESC = 'The path where the environment config is stored.' \
                    ' (default to None to use the provided default environment)'
ARG_PORT_DESC = 'The port to serve grid2viz on.'\
                ' (default to 8050)'
ARG_DEBUG_DESC = 'Enable debug mode for developers.' \
                 ' (default to False)'
ARG_N_CORES_DESC = 'The number of cores to use for the first loading of the best agents of each scenario'

ARG_CACHE_DESC = 'True if you want to build all the cache data for all agents at once before relaunching grid2viz'

def main():
    parser_main = argparse.ArgumentParser(description='Grid2Viz')
    parser_main.add_argument('--agents_path', default=None,
                             required=False, type=str,
                             help=ARG_AGENTS_PATH_DESC)
    parser_main.add_argument('--env_path', default=None,
                             required=False, type=str,
                             help=ARG_ENV_PATH_DESC)
    parser_main.add_argument('--port', default=8050,
                             required=False, type=int,
                             help=ARG_PORT_DESC)
    parser_main.add_argument('--debug', action='store_true',
                             help=ARG_DEBUG_DESC)

    parser_main.add_argument('--n_cores', default=2, type=int,
                             help=ARG_N_CORES_DESC)
    parser_main.add_argument('--cache', default=False, type=bool,
                             help=ARG_CACHE_DESC)


    args = parser_main.parse_args()
    pkg_root_dir = os.path.dirname(os.path.abspath(__file__))
    main_inner(pkg_root_dir, args.agent_path, args.env_path,
               args.port, args.debug, args.n_cores, args.cache)


def main_inner(pkg_root_dir, agents_path=None, env_path=None, port=8050,
               debug=False, n_cores=2, cache=False):

    os.environ["GRID2VIZ_ROOT"] = pkg_root_dir
    config_path = os.path.join(pkg_root_dir, "config.ini")

    if agents_path is not None:
        agents_dir = os.path.abspath(agents_path)
    else:            
        agents_dir = os.path.join(pkg_root_dir, "data", "agents")
        print("Using default agents logs at {}".format(agents_dir))

    if env_path is not None:
        env_dir = os.path.abspath(env_path)
    else:
        env_dir = os.path.join(pkg_root_dir, "data", "env_conf")
        print("Using default environment at {}".format(env_dir))

    with open(config_path, "w") as f:
        f.write(CONFIG_FILE_CONTENT.format(agents_dir=agents_dir,
                                           env_dir=env_dir,
                                           n_cores=n_cores))

    is_makeCache_only=cache

    # Inline import to load app only now
    if(is_makeCache_only):
        make_cache()
    else:
        from grid2viz.app import app_run
        app_run(port, debug)


def make_cache():
    from grid2viz.src.manager import (scenarios, agents,
                                      make_episode_without_decorate,
                                      n_cores, retrieve_episode_from_disk,
                                      save_in_ram_cache, cache_dir)

    from pathos.multiprocessing import ProcessPool
    if not os.path.exists(cache_dir):
        print('Starting Multiprocessing for reading the best agent of each scenario')

    # TODO: tous les agents n'ont pas forcément tourner sur exactement tous les mêmes scenarios
    # Eviter une erreur si un agent n'a pas tourné sur un scenario
    agent_scenario_list=[(agent,scenario) for agent in agents for scenario in scenarios]

    agents_data=[]
    if(n_cores==1):#no multiprocess useful for debug if needed
        i = 0
        for agent_scenario in agent_scenario_list:
            agents_data[i]=make_episode_without_decorate(agent_scenario[0],agent_scenario[1])
            i+=1
    else:
        pool = ProcessPool(n_cores)
        agents_data = list(pool.imap(make_episode_without_decorate,
              [agent_scenario[0] for agent_scenario in agent_scenario_list],#agents
              [agent_scenario[1] for agent_scenario in agent_scenario_list]))#scenarios #we go over all agents and all scenarios for each agent
        pool.close()
        print('Multiprocessing done')

    #####
    #saving data on disk
    i=0
    for agent_scenario in agent_scenario_list:
        print(i)
        agent=agent_scenario[0]
        episode_name=agent_scenario[1]
        agent_episode = agents_data[i]
        if agent_episode is not None:
            episode_data = retrieve_episode_from_disk(
                agent_episode.episode_name, agent_episode.agent)

            agent_episode.decorate(episode_data)
            save_in_ram_cache(agent_episode.episode_name,
                                  agent_episode.agent,
                                  agent_episode)
        i+=1


if __name__ == "__main__":
    main_inner(pkg_root_dir=os.path.dirname(os.path.abspath(__file__)),
               agents_path="D:/Projects/RTE-Grid2Viz/grid2viz/NeurIPS_1.2.2_week1_debug",
               n_cores=2)





import numpy as np
import pandas as pd
import plotly.graph_objects as go

from grid2viz.src.utils.graph_utils import layout_no_data, layout_def


def get_action_per_line(new_episode):
    count = get_modified_lines(new_episode)
    return [go.Bar(x=count.index, y=count.values, name=new_episode.agent)]


def get_modified_lines(new_episode):
    data = get_action_table_data(new_episode)
    # Below to flatten the series of lists "lines_modified"
    try:
        s = data[(data["action_line"] > 0)]["lines_modified"].apply(pd.Series).stack()
        count = s.value_counts()
    except (IndexError, AttributeError):
        count = pd.Series(dtype=np.float64)
    return count


def get_action_redispatch(new_epsiode):
    count = get_modified_gens(new_epsiode)
    return [go.Bar(x=count.index, y=count.values, name=new_epsiode.agent)]


def get_modified_gens(new_episode):
    data = get_action_table_data(new_episode)
    # Below to flatten the series of lists "lines_modified"
    try:
        s = data[(data["action_redisp"] > 0)]["gens_modified"].apply(pd.Series).stack()
        count = s.value_counts()
    except (IndexError, AttributeError):
        count = pd.Series(dtype=np.float64)
    return count


def get_action_table_data(new_episode):
    return new_episode.action_data_table


def get_action_per_sub(new_episode):
    data = get_action_table_data(new_episode)
    # Below to flatten the series of lists "subs_modified"
    try:
        s = data[(data["action_subs"] > 0)]["subs_modified"].apply(pd.Series).stack()
        count = s.value_counts()
    except (IndexError, AttributeError):
        count = pd.Series(dtype=np.float64)
    return [go.Bar(x=count.index, y=count.values, name=new_episode.agent)]


def update_layout(predicate, msg):
    if predicate:
        figure_layout = layout_no_data(msg)
    else:
        figure_layout = layout_def

    return figure_layout


def get_actions_sum(new_episode):
    return (
        new_episode.action_data_table.set_index("timestamp")[
            ["action_line", "action_subs", "action_redisp"]
        ]
        .sum(axis=1)
        .to_frame(name="Nb Actions")
    )



import dash_bootstrap_components as dbc
import dash_core_components as dcc
import dash_html_components as html

nav_items = [
    dbc.NavItem(
        dbc.NavLink(
            "Scenario Selection", active=True, href="/episodes", id="nav_scen_select"
        )
    ),
    dbc.NavItem(dbc.NavLink("Scenario Overview", href="/overview", id="nav_scen_over")),
    dbc.NavItem(dbc.NavLink("Agent Overview", href="/macro", id="nav_agent_over")),
    dbc.NavItem(dbc.NavLink("Agent Study", href="/micro", id="nav_agent_study")),
    dbc.DropdownMenu(
        label="Help",
        color="link",
        in_navbar=True,
        nav=True,
        right=True,
        children=[
            dbc.DropdownMenuItem("Page Help", id="page_help"),
            dbc.DropdownMenuItem(divider=True),
            dbc.NavLink(
                "Documentation",
                href="https://grid2viz.readthedocs.io/en/latest/",
                id="nav_help",
                target="_blank",
            ),
        ],
    ),
]

navbar = dbc.Navbar(
    [
        html.Div(
            [
                html.Span("Scenario:", className="badge badge-secondary"),
                html.Span("", className="badge badge-light", id="scen_lbl"),
            ],
            className="reminder float-left",
        ),
        html.Div(
            [
                html.Span("Ref Agent:", className="badge badge-secondary"),
                html.Span(
                    children=[
                        dbc.Select(
                            id="select_ref_agent",
                            bs_size="sm",
                            disabled=True,
                            placeholder="Ref Agent",
                        )
                    ],
                    className="badge",
                ),
            ],
            className="reminder float-left",
        ),
        html.Div(
            [
                html.Span("Studied Agent:", className="badge badge-secondary"),
                html.Span(
                    children=[
                        dbc.Select(
                            id="select_study_agent",
                            bs_size="sm",
                            disabled=True,
                            placeholder="Study Agent",
                        )
                    ],
                    className="badge",
                ),
            ],
            className="reminder float-left",
        ),
        html.Div(
            [
                html.Div(
                    [
                        dbc.Button(
                            id="enlarge_left",
                            children="-5",
                            color="dark",
                            className="float-left mr-1",
                        ),
                        dbc.Tooltip(
                            "Enlarge left", target="enlarge_left", placement="bottom"
                        ),
                    ]
                ),
                dcc.Dropdown(
                    id="user_timestamps", className="", style={"width": "200px"}
                ),
                html.Div(
                    [
                        dbc.Button(
                            id="enlarge_right",
                            children="+5",
                            color="dark",
                            className="float-left ml-1",
                        ),
                        dbc.Tooltip(
                            "Enlarge right", target="enlarge_right", placement="bottom"
                        ),
                    ]
                ),
            ],
            id="user_timestamp_div",
            className="col-xl-1",
        ),
        html.Div(dbc.Nav(nav_items, navbar=True, pills=True), className="nav_menu"),
    ],
    color="#2196F3",
    sticky="top",
    dark=True,
)

body = html.Div(
    [
        dcc.Location(id="url", refresh=False),
        html.Div(id="page-content", className="main-container"),
        dbc.Modal(
            [
                dbc.ModalBody(
                    [
                        dbc.Spinner(color="primary", type="grow"),
                        html.P("loading graph..."),
                    ]
                )
            ],
            id="loading_modal",
        ),
    ]
)


def make_layout(app):
    app.layout = html.Div(
        [
            dcc.Store(id="scenario", storage_type="memory"),
            dcc.Store(id="agent_ref", storage_type="memory"),
            dcc.Store(id="agent_study", storage_type="memory"),
            dcc.Store(id="user_timestamps_store"),
            dcc.Store(id="window"),
            dcc.Store(id="page"),
            dcc.Store(id="relayoutStoreMicro"),
            dcc.Store(id="reset_timeseries_table_macro", data=True),
            navbar,
            body,
        ]
    )



import numpy as np
import pandas as pd
import plotly.graph_objects as go

from . import observation_model
from .env_actions import env_actions

# colors for production share sunburst pie
dic_colors_prod_types = {
    "nuclear": "darkgoldenrod",
    "thermal": "coral",
    "wind": "darkgreen",
    "solar": "gold",
    "hydro": "steelblue",
}
##WARNING: make sure the "light" color equivalent exists or override it with another "light" color name
dic_light_colors_prod_types = {k: "light" + v for k, v in dic_colors_prod_types.items()}
dic_light_colors_prod_types["nuclear"] = "goldenrod"
dic_light_colors_prod_types["wind"] = "green"
dic_light_colors_prod_types["solar"] = "palegoldenrod"


def get_total_overflow_trace(episode_analytics, episode_data):
    df = get_total_overflow_ts(episode_analytics, episode_data)
    # line_in_overflow=
    return [
        go.Scatter(
            x=df["time"],
            y=df["value"],
            text=[
                "lines " + str(episode_data.line_names[liste]) if len(liste) > 0 else ""
                for liste in df["line_ids"]
            ],  # could be improve with names maybe, here only ids
            name="Nb of overflows",
        )
    ]


def get_total_overflow_ts(episode_analytics, episode_data):
    df = pd.DataFrame(
        index=episode_analytics.timesteps, columns=["time", "value", "line_ids"]
    )
    for (time_step, obs) in enumerate(episode_data.observations):
        # TODO: observations length and timsteps length should match
        try:
            tstamp = episode_analytics.timestamps[time_step]
            ov = obs.timestep_overflow
            overflows = [i for i in range(len(ov)) if ov[i] == 1]
            df.loc[time_step, :] = [tstamp, (ov > 0).sum(), overflows]
        except:
            pass
    return df


def get_prod_share_trace(episode):
    prod_types = episode.get_prod_types()
    prod_type_values = list(prod_types.values()) if len(prod_types.values()) > 0 else []

    share_prod = observation_model.get_prod(episode)
    df = share_prod.groupby("equipment_name")["value"].sum()
    unique_prod_types = np.unique(prod_type_values)

    labels = [*df.index.values, *np.unique(prod_type_values)]

    parents = [prod_types.get(name) for name in df.index.values]
    # labelTypes=[l if l in dic_colors_prod_types.keys() else prod_types.get(l) for l in labels]
    colors = [
        dic_colors_prod_types[l]
        if l in dic_colors_prod_types.keys()
        else dic_light_colors_prod_types[prod_types.get(l)]
        for l in labels
    ]
    values = list(df)

    for prod_type in unique_prod_types:
        parents.append("")
        value = 0
        for gen in df.index.values:
            if prod_types.get(gen) == prod_type:
                value = value + df.get(gen)
        values.append(value)

    return [
        go.Sunburst(
            labels=labels,
            values=values,
            parents=parents,
            branchvalues="total",
            marker=dict(colors=colors, colorscale="RdBu"),
        )
    ]


def get_hazard_trace(episode, equipments=None):
    ts_hazards_by_line = env_actions(episode, which="hazards", kind="ts", aggr=False)

    if ts_hazards_by_line.empty:
        return []

    if "total" in equipments:
        ts_hazards_by_line = ts_hazards_by_line.assign(
            total=episode.hazards.groupby("timestamp", as_index=True)["value"].sum()
        )

    if equipments is not None:
        ts_hazards_by_line = ts_hazards_by_line.loc[:, equipments]

    return [
        go.Scatter(x=ts_hazards_by_line.index, y=ts_hazards_by_line[line], name=line)
        for line in ts_hazards_by_line.columns
    ]


def get_maintenance_trace(episode, equipments=None):
    ts_maintenances_by_line = env_actions(
        episode, which="maintenances", kind="ts", aggr=False
    )

    if ts_maintenances_by_line.empty:
        return []

    # if equipments is not None:
    #   ts_maintenances_by_line = ts_maintenances_by_line.loc[:, equipments]
    linesEquipment = [
        line for line in equipments if line in ts_maintenances_by_line.columns
    ]
    traces = [
        go.Scatter(
            x=ts_maintenances_by_line.index, y=ts_maintenances_by_line[line], name=line
        )
        for line in linesEquipment
    ]

    if "total" in equipments:
        total_maintenance = ts_maintenances_by_line.sum(axis=1)
        names_maintenace = [
            "lines"
            + str(
                list(
                    episode.line_names[
                        ts_maintenances_by_line[episode.line_names].iloc[i].astype(bool)
                    ]
                )
            )
            if total_maintenance[i] >= 1
            else ""
            for i in range(len(total_maintenance))
        ]

        traces.append(
            go.Scatter(
                x=ts_maintenances_by_line.index,
                y=total_maintenance,
                text=names_maintenace,
                name="total",
            )
        )

    return traces


def get_all_prod_trace(episode, prod_types, selection):
    prod_with_type = observation_model.get_prod(episode).assign(
        prod_type=[
            prod_types.get(equipment_name)
            for equipment_name in observation_model.get_prod(episode)["equipment_name"]
        ]
    )
    prod_type_names = prod_types.values()
    trace = []
    if "total" in selection:
        total_series = prod_with_type.groupby("timestamp")["value"].sum()
        trace.append(go.Scatter(x=total_series.index, y=total_series, name="total"))
    for name in prod_type_names:
        if name in selection:
            trace.append(
                go.Scatter(
                    x=prod_with_type[prod_with_type.prod_type.values == name][
                        "timestamp"
                    ].drop_duplicates(),
                    y=prod_with_type[prod_with_type.prod_type.values == name]
                    .groupby(["timestamp"])["value"]
                    .sum(),
                    name=name,
                )
            )
            selection.remove(
                name
            )  # remove prod type from selection to avoid misunderstanding in get_def_trace_per_equipment()

    return [
        *trace,
        *get_df_trace_per_equipment(observation_model.get_prod(episode, selection)),
    ]


def get_load_trace_per_equipment(episode, equipements):
    all_equipements = observation_model.get_load(episode)
    load_equipments = observation_model.get_load(episode, equipements)

    if "total" in equipements:
        # filter intercos
        idx_no_interco = [
            i
            for i in range(len(all_equipements.equipment_name))
            if "interco" not in all_equipements.equipment_name[i]
        ]

        load_equipments = load_equipments.append(
            pd.DataFrame(
                {
                    "equipement_id": [
                        "nan"
                        for i in all_equipements.loc[idx_no_interco]
                        .groupby("timestep")
                        .size()
                    ],
                    "equipment_name": [
                        "total"
                        for i in all_equipements.loc[idx_no_interco]
                        .groupby("timestep")
                        .size()
                    ],
                    "timestamp": [
                        timestamp
                        for timestamp in all_equipements.loc[idx_no_interco][
                            "timestamp"
                        ].unique()
                    ],
                    "timestep": [
                        timestep
                        for timestep in all_equipements.loc[idx_no_interco][
                            "timestep"
                        ].unique()
                    ],
                    "value": [
                        value
                        for value in all_equipements.loc[idx_no_interco]
                        .groupby("timestep")["value"]
                        .sum()
                    ],
                }
            )
        )
    if "total_intercos" in equipements:
        # filter intercos
        idx_interco = [
            i
            for i in range(len(all_equipements.equipment_name))
            if "interco" in all_equipements.equipment_name[i]
        ]
        if len(idx_interco) != 0:
            load_equipments = load_equipments.append(
                pd.DataFrame(
                    {
                        "equipement_id": [
                            "nan"
                            for i in all_equipements.loc[idx_interco]
                            .groupby("timestep")
                            .size()
                        ],
                        "equipment_name": [
                            "total"
                            for i in all_equipements.loc[idx_interco]
                            .groupby("timestep")
                            .size()
                        ],
                        "timestamp": [
                            timestamp
                            for timestamp in all_equipements.loc[idx_interco][
                                "timestamp"
                            ].unique()
                        ],
                        "timestep": [
                            timestep
                            for timestep in all_equipements.loc[idx_interco][
                                "timestep"
                            ].unique()
                        ],
                        "value": [
                            value
                            for value in all_equipements.loc[idx_interco]
                            .groupby("timestep")["value"]
                            .sum()
                        ],
                    }
                )
            )
        else:
            load_equipments = load_equipments.append(
                pd.DataFrame(
                    {
                        "equipement_id": [
                            "nan" for i in all_equipements.groupby("timestep").size()
                        ],
                        "equipment_name": [
                            "total" for i in all_equipements.groupby("timestep").size()
                        ],
                        "timestamp": [
                            timestamp
                            for timestamp in all_equipements["timestamp"].unique()
                        ],
                        "timestep": [
                            timestep
                            for timestep in all_equipements["timestep"].unique()
                        ],
                        "value": [
                            0 for timestep in all_equipements["timestep"].unique()
                        ],
                    }
                )
            )

    return get_df_trace_per_equipment(load_equipments)


def get_usage_rate_trace(episode):
    df = observation_model.get_usage_rate(episode)
    line = {"shape": "spline", "width": 0, "smoothing": 1}
    trace = [
        go.Scatter(
            x=df["timestamp"],
            y=df["value"]["quantile10"],
            name="quantile 10",
            line=line,
        ),
        go.Scatter(
            x=df["timestamp"],
            y=df["value"]["quantile25"],
            name="quantile 25",
            fill="tonexty",
            fillcolor="rgba(159, 197, 232, 0.63)",
            line=line,
        ),
        go.Scatter(
            x=df["timestamp"],
            y=df["value"]["median"],
            name="median",
            fill="tonexty",
            fillcolor="rgba(31, 119, 180, 0.5)",
            line={"color": "rgb(31, 119, 180)", "shape": "spline", "smoothing": 1},
        ),
        go.Scatter(
            x=df["timestamp"],
            y=df["value"]["quantile75"],
            name="quantile 75",
            fill="tonexty",
            fillcolor="rgba(31, 119, 180, 0.5)",
            line=line,
        ),
        go.Scatter(
            x=df["timestamp"],
            y=df["value"]["quantile90"],
            name="quantile 90",
            fill="tonexty",
            fillcolor="rgba(159, 197, 232, 0.63)",
            line=line,
        ),
        go.Scatter(
            x=df["timestamp"],
            y=df["value"]["max"],
            name="Max",
            line={"shape": "spline", "smoothing": 1, "color": "rgba(255,0,0,0.5)"},
        ),
    ]
    return trace


def get_df_trace_per_equipment(df):
    return [
        go.Scatter(
            x=df.loc[df["equipment_name"] == equipment, :]["timestamp"],
            y=df.loc[df["equipment_name"] == equipment, :]["value"],
            name=equipment,
        )
        for equipment in df["equipment_name"].drop_duplicates()
    ]


def get_df_rewards_trace(episode):
    df = observation_model.get_df_computed_reward(episode)
    return [
        go.Scatter(x=df["timestep"], y=df["rewards"], name=episode.agent + "_reward"),
        go.Scatter(
            x=df["timestep"],
            y=df["cum_rewards"],
            name=episode.agent + "cum_rewards",
            yaxis="y2",
        ),
    ]


def get_attacks_trace(episode, equipments=None):
    df = episode.attacks_data_table
    return [
        go.Scatter(
            x=df["timestamp"],
            y=df["attack"].astype(int),
            name=episode.agent + "_attacks",
            text="lines " + df["id_lines"],
        )
    ]



"""
This file handles the html entry point of the application through dash components.
It will generate the layout of a given page and handle the routing
"""

import dash_bootstrap_components as dbc

# from dash import Dash
from jupyter_dash import JupyterDash

# We need to create app before importing the rest of the project as it uses @app decorators
JupyterDash.infer_jupyter_proxy_config()  # for binder or jupyterHub for instance
app = JupyterDash(
    __name__, external_stylesheets=[dbc.themes.BOOTSTRAP]
)  # ,server_url="http://127.0.0.1:8050/")

"""
Get Imports to create layout and callbacks 
"""
from grid2viz.main_callbacks import register_callbacks_main
from grid2viz.layout import make_layout as layout

from grid2viz.src.episodes.episodes_clbk import register_callbacks_episodes
from grid2viz.src.overview.overview_clbk import (
    register_callbacks_overview,
)  # as overview_clbk
from grid2viz.src.macro.macro_clbk import register_callbacks_macro  # as macro_clbk
from grid2viz.src.micro.micro_clbk import register_callbacks_micro  # as micro_clbk

"""
End Warning
"""

app.config.suppress_callback_exceptions = True

app.title = "Grid2Viz"
app.server.secret_key = "Grid2Viz"

##create layout
layout(app)

##create callbaks
register_callbacks_main(app)
register_callbacks_episodes(app)
register_callbacks_overview(app)
register_callbacks_macro(app)
register_callbacks_micro(app)


def app_run(port=8050, debug=False):
    app.run_server(port=port, debug=debug)


if __name__ == "__main__":
    app_run()



from pathlib import Path

import dash_antd_components as dac
import dash_bootstrap_components as dbc
import dash_core_components as dcc
import dash_html_components as html
import plotly.figure_factory as ff
import seaborn as sn

from grid2viz.src.manager import survival_df, grid2viz_home_directory, scenarios
from grid2viz.src.utils.constants import DONT_SHOW_FILENAME
from grid2viz.src.utils.layout_helpers import modal, should_help_open


# This is to improve readability of the heatmap of survival steps for agents
def get_heatmap_survival_score(df):
    if (df.shape[0] >= 2) and (df.shape[1] >= 2):
        clustered_df = sn.clustermap(df)
        reordered_scenarios = clustered_df.dendrogram_row.reordered_ind
        reordered_agents = clustered_df.dendrogram_col.reordered_ind
        return df.iloc[reordered_scenarios, reordered_agents]
    else:
        return df


def create_heatmap_figure(df):
    clustered_survival_df = get_heatmap_survival_score(df)

    z_text = clustered_survival_df.copy().astype(str)
    z_text[z_text == "-1"] = ""

    heatmap_figure = ff.create_annotated_heatmap(  # go.Figure(data=go.Heatmap(
        z=clustered_survival_df.values,  # survival_df.values,#z=pd.concat([survival_df, survival_df]))),
        x=clustered_survival_df.columns.tolist(),
        y=clustered_survival_df.index.tolist(),
        colorscale="RdYlGn",
        zmid=50,
        annotation_text=z_text.values,
    )
    heatmap_figure.update_layout(
        {"yaxis": {"type": "category"}, "xaxis": {"type": "category"}}
    )
    return heatmap_figure


def generate_heatmap_components(df):
    heatmap_div = html.Div(
        children=[
            html.H5("Percentage of Agents' survival time over a scenario"),
            dcc.Graph(
                id="heatmap",
                figure=create_heatmap_figure(df),
            ),
        ],
        className="col-xl-12 align-self-center heatmap",
    )

    return html.Div(
        className="lineBlockSlim card",
        children=[
            dbc.Collapse(heatmap_div, id="collapse"),
            scenarios_filter(sorted(list(scenarios))),
        ],
    )


def scenarios_filter(scenarios):
    return html.Div(
        id="scenario_filter_div",
        children=[
            html.H5("Select the scenarios you want to see below."),
            dac.Select(
                id="scenarios_filter",
                options=[
                    {"label": scenario, "value": scenario} for scenario in scenarios
                ],
                value=scenarios,
                mode="multiple",
            ),
        ],
    )


def comparison_button():
    return html.Div(
        className="row comparison-btn",
        children=[
            dbc.Button(
                "Open scenarios comparison & filtering",
                id="collapse-button",
                color="info",
                size="lg",
                outline=True,
                block=True,
            )
        ],
    )


def layout():
    open_help = should_help_open(
        Path(grid2viz_home_directory) / DONT_SHOW_FILENAME("episodes")
    )
    header = "Take a look at the Scenarios"
    body = (
        " Choose a specific scenario to analyze by clicking its Open button.  "
        "Get more detailed information when choosing your scenario by clicking on the comparison & filtering button"
    )
    return html.Div(
        id="scenario_page",
        children=[
            dcc.Store(id="relayoutStoreScenario"),
            comparison_button(),
            generate_heatmap_components(survival_df),
            dbc.Row(id="cards_container", className="m-1"),
            modal(id_suffix="episodes", is_open=open_help, header=header, body=body),
        ],
    )



from pathlib import Path

import pandas as pd
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

from grid2viz.src.kpi import observation_model, EpisodeTrace
from grid2viz.src.manager import make_episode, best_agents, grid2viz_home_directory
from grid2viz.src.utils import common_graph
from grid2viz.src.utils.callbacks_helpers import toggle_modal_helper
from grid2viz.src.utils.constants import DONT_SHOW_FILENAME
from grid2viz.src.utils.graph_utils import relayout_callback, get_axis_relayout


def register_callbacks_overview(app):
    @app.callback(
        Output("relayoutStoreOverview", "data"),
        [
            Input("input_env_charts", "relayoutData"),
            Input("usage_rate_graph", "relayoutData"),
            Input("overflow_graph", "relayoutData"),
        ],
        [State("relayoutStoreOverview", "data")],
    )
    def relayout_store_overview(*args):
        return relayout_callback(*args)

    @app.callback(
        [
            Output("input_assets_selector", "options"),
            Output("input_assets_selector", "value"),
        ],
        [Input("scen_overview_ts_switch", "value")],
        [State("scenario", "data")],
    )
    def update_ts_graph_avail_assets(kind, scenario):
        """
        Change the selector's options according to the kind of trace selected.

        Triggered when user click on one of the input in the scen_overview_ts_switch
        component in overview layout.
        """
        best_agent_ep = make_episode(best_agents[scenario]["agent"], scenario)
        return common_graph.ts_graph_avail_assets(kind, best_agent_ep)

    @app.callback(
        Output("input_env_charts", "figure"),
        [
            Input("input_assets_selector", "value"),
            Input("relayoutStoreOverview", "data"),
        ],
        [
            State("input_env_charts", "figure"),
            State("scen_overview_ts_switch", "value"),
            State("scenario", "data"),
        ],
    )
    def load_environments_ts(equipments, relayout_data_store, figure, kind, scenario):
        """
        Load selected kind of environment for chosen equipments in a scenario.

        Triggered when user click on a equipment displayed in the
        input_assets_selector in the overview layout.
        """
        if relayout_data_store is not None and relayout_data_store["relayout_data"]:
            relayout_data = relayout_data_store["relayout_data"]
            layout = figure["layout"]
            new_axis_layout = get_axis_relayout(figure, relayout_data)
            if new_axis_layout is not None:
                layout.update(new_axis_layout)
                return figure

        if kind is None:
            return figure
        if isinstance(equipments, str):
            equipments = [equipments]  # to make pd.series.isin() work

        figure["data"] = common_graph.environment_ts_data(
            kind, make_episode(best_agents[scenario]["agent"], scenario), equipments
        )

        return figure

    @app.callback(
        Output("select_loads_for_tb", "options"),
        [Input("indicator_line", "children")],
        [State("scenario", "data")],
    )
    def update_select_loads(children, scenario):
        """
        Display list of loads in a the select_loads_for_tb component.

        Triggered when indicator line is loaded.
        """
        episode = make_episode(best_agents[scenario]["agent"], scenario)
        return [
            {"label": load, "value": load}
            for load in [*episode.load_names, "total", "total_intercos"]
        ]

    @app.callback(
        Output("select_prods_for_tb", "options"),
        [Input("indicator_line", "children")],
        [State("scenario", "data")],
    )
    def update_select_prods(children, scenario):
        """
        Display list of production in a the select_prods_for_tb component.

        Triggered when indicator line is loaded.
        """
        episode = make_episode(best_agents[scenario]["agent"], scenario)
        return [{"label": prod, "value": prod} for prod in episode.prod_names]

    @app.callback(
        [Output("inspection_table", "columns"), Output("inspection_table", "data")],
        [
            Input("select_loads_for_tb", "value"),
            Input("select_prods_for_tb", "value"),
            Input("agent_ref", "data"),
            Input("date_range", "start_date"),
            Input("date_range", "end_date"),
        ],
        [State("inspection_table", "data"), State("scenario", "data")],
    )
    def update_table(loads, prods, agent_ref, start_date, end_date, data, scenario):
        """
        Update the inspection table with the loads and prods selected.

        Triggered when the select a load or a prods and when the ref agent is changed.
        """
        if agent_ref is None:
            raise PreventUpdate
        episode = make_episode(agent_ref, scenario)
        df = observation_model.init_table_inspection_data(episode)
        if data is None:
            return [{"name": i, "id": i} for i in df.columns], df.to_dict("records")
        if loads is None:
            loads = []
        if prods is None:
            prods = []
        df["timestamp"] = pd.to_datetime(df["timestamp"])
        cols_to_drop = []
        for col in df.columns[4:]:
            if col not in loads and col not in prods:
                cols_to_drop.append(col)
        cols_to_add = [col for col in loads + prods if col not in df.columns]
        df = df.drop(cols_to_drop, axis=1)
        if cols_to_add:
            df = df.merge(
                observation_model.get_prod_and_conso(episode)[cols_to_add],
                left_on="timestamp",
                right_index=True,
                how="right",
            )
        if start_date is not None:
            df = df[df["timestamp"] >= start_date]
        if end_date is not None:
            df = df[df["timestamp"] <= end_date]
        cols = [{"name": i, "id": i} for i in df.columns]
        return cols, df.to_dict("records")

    @app.callback(Output("nb_steps_card", "children"), [Input("scenario", "data")])
    def update_card_step(scenario):
        """Display the best agent number of step when the page is loaded."""
        best_agent_ep = make_episode(best_agents[scenario]["agent"], scenario)
        return "{} / {}".format(
            best_agent_ep.meta["nb_timestep_played"],
            best_agent_ep.meta["chronics_max_timestep"],
        )

    @app.callback(
        Output("nb_maintenance_card", "children"), [Input("scenario", "data")]
    )
    def update_card_maintenance(scenario):
        """Display the number of maintenance of the best agent when page is loaded."""
        best_agent_ep = make_episode(best_agents[scenario]["agent"], scenario)
        return best_agent_ep.nb_maintenances

    @app.callback(Output("nb_hazard_card", "children"), [Input("scenario", "data")])
    def update_card_hazard(scenario):
        """Display the number of hazard of the best agent when page is loaded."""
        best_agent_ep = make_episode(best_agents[scenario]["agent"], scenario)
        return best_agent_ep.nb_hazards

    @app.callback(
        Output("duration_maintenance_card", "children"), [Input("scenario", "data")]
    )
    def update_card_duration_maintenances(scenario):
        """
        Display the total duration of maintenances made by the best agent when
        page is loaded.
        """
        best_agent_ep = make_episode(best_agents[scenario]["agent"], scenario)
        return best_agent_ep.total_maintenance_duration

    # @app.callback(
    #     Output("agent_ref", "data"),
    #     [Input("input_agent_selector", "value")],
    #     [State("scenario", "data")]
    # )
    # def update_selected_ref_agent(ref_agent, scenario):
    #     """
    #         Change the agent of reference for the given scenario.
    #
    #         Triggered when user select a new agent with the agent selector on layout.
    #     """
    #     make_episode(ref_agent, scenario)
    #     return ref_agent

    @app.callback(
        [Output("overflow_graph", "figure"), Output("usage_rate_graph", "figure")],
        [
            Input("agent_ref", "data"),
            Input("scenario", "data"),
            Input("relayoutStoreOverview", "data"),
        ],
        [State("overflow_graph", "figure"), State("usage_rate_graph", "figure")],
    )
    def update_agent_ref_graph(
        ref_agent, scenario, relayout_data_store, figure_overflow, figure_usage
    ):
        if relayout_data_store is not None and relayout_data_store["relayout_data"]:
            relayout_data = relayout_data_store["relayout_data"]
            layout_usage = figure_usage["layout"]
            new_axis_layout = get_axis_relayout(figure_usage, relayout_data)
            if new_axis_layout is not None:
                layout_usage.update(new_axis_layout)
                figure_overflow["layout"].update(new_axis_layout)
                return figure_overflow, figure_usage

        return common_graph.agent_overflow_usage_rate_trace(
            make_episode(ref_agent, scenario), figure_overflow, figure_usage
        )

    @app.callback(
        Output("indicator_line_charts", "figure"),
        [Input("scenario", "data")],
        [State("indicator_line_charts", "figure")],
    )
    def update_profile_conso_graph(scenario, figure):
        """Display best agent's consumption profile when page is loaded"""
        best_agent_ep = make_episode(best_agents[scenario]["agent"], scenario)
        figure["data"] = best_agent_ep.profile_traces
        return figure

    @app.callback(
        Output("production_share_graph", "figure"),
        [Input("scenario", "data")],
        [State("production_share_graph", "figure")],
    )
    def update_production_share_graph(scenario, figure):
        """Display best agent's production share when page load"""
        best_agent_ep = make_episode(best_agents[scenario]["agent"], scenario)
        share_prod = EpisodeTrace.get_prod_share_trace(best_agent_ep)
        figure["data"] = share_prod
        return figure

    @app.callback(
        [Output("date_range", "start_date"), Output("date_range", "end_date")],
        [Input("agent_ref", "data")],
        [State("scenario", "data")],
    )
    def update_date_range(agent_ref, scenario):
        """Change the date range for the date picker in inspector line"""
        episode = make_episode(agent_ref, scenario)
        return (
            episode.production["timestamp"].dt.date.values[0],
            episode.production["timestamp"].dt.date.values[-1],
        )

    @app.callback(
        [
            Output("modal_overview", "is_open"),
            Output("dont_show_again_div_overview", "className"),
        ],
        [Input("close_overview", "n_clicks"), Input("page_help", "n_clicks")],
        [
            State("modal_overview", "is_open"),
            State("dont_show_again_overview", "checked"),
        ],
    )
    def toggle_modal(close_n_clicks, open_n_clicks, is_open, dont_show_again):
        dsa_filepath = Path(grid2viz_home_directory) / DONT_SHOW_FILENAME("overview")
        return toggle_modal_helper(
            close_n_clicks,
            open_n_clicks,
            is_open,
            dont_show_again,
            dsa_filepath,
            "page_help",
        )

    @app.callback(Output("modal_image_overview", "src"), [Input("url", "pathname")])
    def show_image(pathname):
        return app.get_asset_url("screenshots/scenario_overview.png")



import os
import pathlib
import unittest

# We need to make this below so that the manager.py finds the config.ini
os.environ['GRID2VIZ_ROOT'] = os.path.join(
            pathlib.Path(__file__).parent.absolute(), 'data')

agents_path = os.path.join(
            pathlib.Path(__file__).parent.absolute(), 'data', 'agents')

config_str = f'[DEFAULT]\nagents_dir={agents_path}'
config_file_path = os.path.join(os.environ['GRID2VIZ_ROOT'], 'config.ini')

with open(config_file_path, 'w') as f:
    f.write(config_str)

from grid2op import make
from grid2op.Backend import PandaPowerBackend
from grid2op.Episode.EpisodeData import EpisodeData
from grid2op.Parameters import Parameters
from grid2op.Runner import Runner

from CustomAgent import RandomRedispatchAgent
from grid2viz.src.kpi.EpisodeAnalytics import EpisodeAnalytics


class TestGenerateAgent(unittest.TestCase):
    def setUp(self):
        self.case = "rte_case14_realistic"
        self.backend = PandaPowerBackend()
        self.param = Parameters()

        self.agents_path = agents_path
        self.agent_name = 'redispatching-baseline'
        self.scenario_name = '000'

    def test_generate_and_read_agent_redisp(self):
        with make(self.case, param=self.param, backend=self.backend) as env:
            agent = RandomRedispatchAgent(env.action_space, env)
            runner = Runner(**env.get_params_for_runner(),
                            agentClass=None,
                            agentInstance=agent)
            # need to be seeded for reproducibility as this takes random redispatching actions
            runner.run(nb_episode=1,
                       path_save=os.path.join(self.agents_path, self.agent_name),
                       nb_process=1,
                       max_iter=10,
                       env_seeds=[0],
                       agent_seeds=[0],
                       pbar=True)
            env.close()

        self.episode_data = EpisodeData.from_disk(
            os.path.join(self.agents_path, self.agent_name), self.scenario_name
        )
        self.episode_analytics = EpisodeAnalytics(
            self.episode_data, self.scenario_name, self.agent_name
        )




"""
This file builds the layout for the scenario overview tab.
This tab handles the generic information about the environment and the selection of a reference agent for future analysis.
"""
import base64
import io
from pathlib import Path

import dash_antd_components as dac
import dash_core_components as dcc
import dash_html_components as html
import dash_table as dt
import matplotlib.pyplot as plt
import plotly.graph_objects as go

from grid2viz.src.manager import (
    agent_scenario,
    make_episode,
    best_agents,
    grid2viz_home_directory,
    make_network_scenario_overview,
)
from grid2viz.src.utils.constants import DONT_SHOW_FILENAME
from grid2viz.src.utils.layout_helpers import modal, should_help_open

layout_def = {
    "legend": {"orientation": "h"},
    "showlegend": True,
    "margin": {"l": 0, "r": 0, "t": 0, "b": 0},
}

layout_pie = {
    "legend": {"orientation": "h"},
    "margin": {"l": 0, "r": 0, "t": 0, "b": 0},
    "paper_bgcolor": "rgba(0,0,0,0)",
    "plot_bgcolor": "rgba(0,0,0,0)",
}


def indicators_line(encoded_image):
    return html.Div(
        id="indicator_line",
        children=[
            html.H4("Indicators"),
            html.Div(
                children=[
                    html.Div(
                        [
                            html.H5("Distribution of Daily Load Profiles (MW)"),
                            dcc.Graph(
                                id="indicator_line_charts",
                                style={"margin-top": "1em"},
                                figure=go.Figure(layout=layout_def),
                            ),
                        ],
                        className="col-xl-5",
                    ),
                    html.Div(
                        children=[
                            html.H5("Production Shares"),
                            dcc.Graph(
                                id="production_share_graph",
                                figure=go.Figure(layout=layout_pie),
                            ),
                        ],
                        className="col-xl-4",
                    ),
                    # number summary column
                    html.Div(
                        children=[
                            html.Div(
                                className="mb-4",
                                children=[
                                    html.P(
                                        id="nb_steps_card",
                                        className="border-bottom h3 mb-0 text-right",
                                        children="",
                                    ),
                                    html.P(
                                        className="text-muted",
                                        children="Best Agent's Steps (step / nb actions)",
                                    ),
                                ],
                            ),
                            html.Div(
                                className="mb-4",
                                children=[
                                    html.P(
                                        id="nb_hazard_card",
                                        className="border-bottom h3 mb-0 text-right",
                                        children="",
                                    ),
                                    html.P(
                                        className="text-muted",
                                        children="Number of Hazards",
                                    ),
                                ],
                            ),
                            html.Div(
                                className="mb-4",
                                children=[
                                    html.P(
                                        id="nb_maintenance_card",
                                        className="border-bottom h3 mb-0 text-right",
                                        children="",
                                    ),
                                    html.P(
                                        className="text-muted",
                                        children="Number of Maintenances",
                                    ),
                                ],
                            ),
                            html.Div(
                                className="mb-4",
                                children=[
                                    html.P(
                                        id="duration_maintenance_card",
                                        className="border-bottom h3 mb-0 text-right",
                                        children="",
                                    ),
                                    html.P(
                                        className="text-muted",
                                        children="Maintenances Duration (min)",
                                    ),
                                ],
                            ),
                        ],
                        className="col-xl-3 align-self-center",
                    ),
                    html.Div(
                        [
                            html.H5(
                                "Max prod & laod values and dashed lines in maintenance on Power grid",
                                style={"margin-top": "2%"},
                            ),
                            html.Img(
                                src="data:image/png;base64,{}".format(encoded_image)
                            ),
                        ],
                        className="col-xl-12",
                    ),
                ],
                className="card-body row",
            ),
        ],
        className="lineBlock card",
    )


def summary_line(episode, ref_agent, scenario):
    return html.Div(
        children=[
            html.H4("Summary"),
            html.Div(
                children=[
                    html.Div(
                        children=[
                            html.H5("Environment Time Series"),
                            dac.Radio(
                                options=[
                                    {"label": "Production (MW)", "value": "Production"},
                                    {"label": "Load (MW)", "value": "Load"},
                                    {"label": "Hazards", "value": "Hazards"},
                                    {"label": "Maintenances", "value": "Maintenances"},
                                ],
                                value="Production",
                                id="scen_overview_ts_switch",
                                buttonStyle="solid",
                            ),
                            dac.Select(
                                id="input_assets_selector",
                                options=[
                                    {"label": prod_name, "value": prod_name}
                                    for prod_name in episode.prod_names
                                ],
                                value="solar",
                                mode="multiple",
                                showArrow=True,
                            ),
                            dcc.Graph(
                                id="input_env_charts",
                                style={"margin-top": "1em"},
                                figure=go.Figure(layout=layout_def),
                                config=dict(displayModeBar=False),
                            ),
                        ],
                        className="col-xl-5",
                    ),
                    html.Div(
                        children=[
                            html.H5("Reference agent Metrics"),
                            # dcc.Dropdown(
                            #     id="input_agent_selector", placeholder="select a ref agent",
                            #     options=[{'label': agent, 'value': agent}
                            #              for agent in agent_scenario[scenario]],
                            #     value=ref_agent
                            # ),
                            html.Div(
                                children=[
                                    html.Div(
                                        children=[
                                            html.H5(
                                                "Usage rate", className="text-center"
                                            ),
                                            dcc.Graph(
                                                id="usage_rate_graph",
                                                style={"margin-top": "1em"},
                                                figure=go.Figure(layout=layout_def),
                                                config=dict(displayModeBar=False),
                                            ),
                                        ],
                                        className="col-6",
                                    ),
                                    html.Div(
                                        children=[
                                            html.H5(
                                                "Overflow", className="text-center"
                                            ),
                                            dcc.Graph(
                                                id="overflow_graph",
                                                style={"margin-top": "1em"},
                                                figure=go.Figure(layout=layout_def),
                                                config=dict(displayModeBar=False),
                                            ),
                                        ],
                                        className="col-6",
                                    ),
                                ],
                                className="row",
                            ),
                        ],
                        className="col-xl-7",
                    ),
                ],
                className="card-body row",
            ),
        ],
        className="lineBlock card",
    )


ref_agent_line = html.Div(
    children=[
        html.H4("Inspector"),
        html.Div(
            children=[
                html.Div(
                    children=[
                        html.H5("Table Filters Selection"),
                        dcc.DatePickerRange(
                            id="date_range",
                            display_format="MMM Do, YY",
                        ),
                        html.H5("Loads"),
                        dcc.Dropdown(
                            id="select_loads_for_tb",
                            options=[],
                            multi=True,
                            style=dict(marginBottom="1em"),
                        ),
                        html.H5("Generators"),
                        dcc.Dropdown(
                            id="select_prods_for_tb",
                            options=[],
                            multi=True,
                            style=dict(marginBottom="1em"),
                        ),
                    ],
                    className="col-xl-2",
                ),
                html.Div(
                    children=[
                        html.Label(
                            children=[
                                "The documentation for the filtering syntax within the data table header can be found ",
                                html.A(
                                    "here.",
                                    href="https://dash.plot.ly/datatable/filtering",
                                    target="_blank",
                                ),
                            ]
                        ),
                        dt.DataTable(
                            id="inspection_table",
                            filter_action="native",
                            sort_action="native",
                            sort_mode="multi",
                            page_action="native",
                            page_current=0,
                            page_size=20,
                        ),
                    ],
                    className="col-xl-10",
                ),
            ],
            className="col-xl-10 p-2",
        ),
    ],
    className="lineBlock card",
)


def layout(scenario, ref_agent):
    try:
        episode = make_episode(best_agents[scenario]["agent"], scenario)
    except Exception as ex:
        print(ex)
        return
    if ref_agent is None:
        ref_agent = agent_scenario[scenario][0]

    network_graph = make_network_scenario_overview(episode)

    # /!\ As of 2020/10/29 the mpl_to_plotly functions is broken and not maintained
    # It calls a deprecated function of matplotlib.
    # Work around below : insert the image.
    # https://stackoverflow.com/questions/63120058/plotly-mpl-to-plotly-error-spine-object-has-no-attribute-is-frame-like
    buf = io.BytesIO()
    plt.figure(network_graph.number)
    plt.savefig(buf, format="png")
    buf.seek(0)
    encoded_image = base64.b64encode(buf.read())
    buf.close()

    open_help = should_help_open(
        Path(grid2viz_home_directory) / DONT_SHOW_FILENAME("overview")
    )
    header = "Take a deeper look at the Scenario"
    body = (
        "Look at the grid, inspect indicators as well as chronics"
        "Select a reference agent in the dropdown menu to get a sense of flows and overflows over the scenario"
        "When done, move to Agent Overview section. The reference agent will be used there as baseline to compare with"
    )
    return html.Div(
        id="overview_page",
        children=[
            dcc.Store(id="relayoutStoreOverview"),
            indicators_line(encoded_image.decode()),
            summary_line(episode, ref_agent, scenario),
            ref_agent_line,
            modal(id_suffix="overview", is_open=open_help, header=header, body=body),
        ],
    )



from .env_actions import env_actions


def total_duration_maintenance(episode):
    timestep_duration = episode.timestamps[1] - episode.timestamps[0]
    nb_maintenance = env_actions(
        episode, which="maintenances", kind="dur", aggr=False
    ).sum()
    return (timestep_duration * nb_maintenance).total_seconds() / 60.0


def hist_duration_maintenances(episode):
    # Suppose that there is at most one maintenance per line per episode

    return [t for t in episode.observations[0].duration_next_maintenance if t]



"""Utility functions for manipulating plotly figures"""

from dash.exceptions import PreventUpdate


def max_or_zero(x):
    if len(x) == 0:
        return 0
    else:
        return max(x)


class RelayoutX(object):
    """
    Wrapping class for dash core component Graph relayoutData attribute
    to make it hashable.

    Attributes
    ----------
    xmin : float
        xaxis.range[0] attribute of relayoutData (when it exists, None otherwise).
    xmax : float
        xaxis.range[1] attribute of relayoutData (when it exists, None otherwise).

    """

    def __init__(self, relayout_data=None):

        if "xaxis.range[0]" in relayout_data:
            self.xmin, self.xmax = (
                relayout_data["xaxis.range[0]"],
                relayout_data["xaxis.range[1]"],
            )
        else:
            self.xmin, self.xmax = None, None

        self.relayout_data = relayout_data

    def __eq__(self, other):
        return (self.xmin == other.xmin) and (self.xmax == other.xmax)

    def __hash__(self):
        return hash((self.xmin, self.xmax))


def relayout_callback(*args):
    # This function supposes that it's not possible to get the same
    # relayoutData twice.

    if any(arg is None for arg in args[:-1]):
        raise PreventUpdate
    if args[-1] is None:
        relayout_data_store = dict(relayout_history=[], relayout_data=None, reset_nb=0)
    else:
        relayout_data_store = args[-1]

    if all("autosize" in arg for arg in args[:-1]):
        return relayout_data_store

    reset_nb = relayout_data_store["reset_nb"]
    new_reset_nb = 0
    relayouts_indx = []
    relayout_history_x = [
        RelayoutX(rlyt) for rlyt in relayout_data_store["relayout_history"]
    ]
    relayouts_x = [RelayoutX(arg) for arg in args[:-1]]
    for i, arg in enumerate(args[:-1]):
        if (
            ("xaxis.range[0]" in arg or "xaxis.autorange" in arg)
            and (relayouts_x[i] not in relayout_history_x)
            and ("autosize" not in arg)
        ):
            if "xaxis.autorange" in arg and arg["xaxis.autorange"]:
                new_reset_nb = new_reset_nb + 1
            else:
                relayouts_indx.append(i)
    if not relayouts_indx:
        if new_reset_nb > reset_nb:
            # It's a zoom reset
            relayout_data_store["reset_nb"] = reset_nb + 1
            relayout_data_store["relayout_data"] = {"xaxis.autorange": True}
            return relayout_data_store
        else:
            # No relayout events with xaxis effect
            raise PreventUpdate
    if len(relayouts_indx) > 2:
        # This should never happen
        print(relayouts_indx)
        raise Exception

    relayout_data_store["reset_nb"] = new_reset_nb
    relayout_data_store["relayout_history"].append(args[relayouts_indx[0]])
    relayout_data_store["relayout_data"] = args[relayouts_indx[0]]
    return relayout_data_store


def get_axis_relayout(figure, relayout_data):
    layout = figure["layout"]
    template_layout = figure["layout"]["template"]["layout"]
    if "xaxis" in layout:
        xaxis = layout["xaxis"]
    else:
        xaxis = template_layout["xaxis"]
    if "range" not in xaxis and "x" in figure["data"][0]:
        xaxis.update(range=[min(figure["data"][0]["x"]), max(figure["data"][0]["x"])])
    res = {}
    if "xaxis.range[0]" in relayout_data:
        xmin, xmax = relayout_data["xaxis.range[0]"], relayout_data["xaxis.range[1]"]
        if [xmin, xmax] != xaxis["range"]:
            res.update(xaxis=dict(range=[xmin, xmax], autorange=False))
    if "xaxis.autorange" in relayout_data:
        res.update(xaxis=dict(autorange=relayout_data["xaxis.autorange"]))
    if res:
        return res


layout_def = {
    "legend": {"orientation": "h"},
    "margin": {"l": 0, "r": 0, "t": 0, "b": 0},
    "xaxis": {"visible": True},
    "yaxis": {"visible": True, "fixedrange": True, "autorange": True},
    "annotations": [],
    "plot_bgcolor": "#E5ECF6",
}


def layout_no_data(msg):
    return {
        "xaxis": {"visible": False},
        "yaxis": {"visible": False},
        "plot_bgcolor": "rgba(0, 0, 0, 0)",
        "annotations": [
            {
                "text": msg,
                "xref": "paper",
                "yref": "paper",
                "showarrow": False,
                "font": {"size": 28},
            }
        ],
    }



import os
import pathlib
import unittest

# We need to make this below so that the manager.py finds the config.ini
os.environ['GRID2VIZ_ROOT'] = os.path.join(
            pathlib.Path(__file__).parent.absolute(), 'data')

agents_path = os.path.join(
            pathlib.Path(__file__).parent.absolute(), 'data', 'agents')

config_str = f'[DEFAULT]\nagents_dir={agents_path}'
config_file_path = os.path.join(os.environ['GRID2VIZ_ROOT'], 'config.ini')

with open(config_file_path, 'w') as f:
    f.write(config_str)

from grid2op.Episode.EpisodeData import EpisodeData
from grid2viz.src.kpi.EpisodeAnalytics import EpisodeAnalytics
from grid2viz.src.kpi.actions_model import get_action_per_line, get_action_per_sub


class TestEpisodeAnalytics(unittest.TestCase):
    def setUp(self):
        self.agents_path = agents_path
        self.agent_name = 'greedy-baseline'
        self.scenario_name = '000'
        self.episode_data = EpisodeData.from_disk(
            os.path.join(self.agents_path, self.agent_name), self.scenario_name
        )
        self.episode_analytics = EpisodeAnalytics(
            self.episode_data, self.scenario_name, self.agent_name
        )

    def test_action_data_table(self):

        self.assertListEqual(
            self.episode_analytics.action_data_table.action_id[:3].tolist(),
            [0, 1, 2]
        )
        self.assertListEqual(
            self.episode_analytics.action_data_table.action_id[37:40].tolist(),
            [4, 3, 0]
        )
        self.assertListEqual(
            self.episode_analytics.action_data_table.distance[:3].tolist(),
            [1, 2, 3]
        )
        self.assertListEqual(
            self.episode_analytics.action_data_table.distance[37:40].tolist(),
            [2, 2, 3]
        )

    def test_action_repartition(self):
        nb_actions = self.episode_analytics.action_data_table[
            ['action_line', 'action_subs']].sum()
        self.assertEqual(nb_actions.action_line, 0.0)
        self.assertEqual(nb_actions.action_subs, 177.0)

        action_per_line = get_action_per_line(self.episode_analytics)
        action_per_sub = get_action_per_sub(self.episode_analytics)

        self.assertListEqual(
            action_per_line[0].x.tolist(),
            []
        )
        # We need to sort the labels for which values are equal.
        # Otherwise, the output is random.
        self.assertListEqual(
            [*action_per_sub[0].x.tolist()[:-2], *sorted(action_per_sub[0].x.tolist()[-2:])],
            ['sub_9', 'sub_8', 'sub_4', 'sub_12', 'sub_1', 'sub_3']
        )
        self.assertListEqual(
            action_per_sub[0].y.tolist(),
            [106, 45, 12, 10, 2, 2]
        )

    def test_multi_topo(self):
        self.agent_name = 'multiTopology-baseline'
        self.scenario_name = '000'
        self.episode_data = EpisodeData.from_disk(
            os.path.join(self.agents_path, self.agent_name), self.scenario_name
        )
        self.episode_analytics = EpisodeAnalytics(
            self.episode_data, self.scenario_name, self.agent_name
        )

        nb_actions = self.episode_analytics.action_data_table[
            ['action_line', 'action_subs']].sum()
        self.assertEqual(nb_actions.action_line, 25.0)
        self.assertEqual(nb_actions.action_subs, 38.0)

        action_per_line = get_action_per_line(self.episode_analytics)
        action_per_sub = get_action_per_sub(self.episode_analytics)

        # We need to sort the labels for which values are equal.
        # Otherwise, the output is random.
        self.assertListEqual(
            sorted(action_per_sub[0].x.tolist()),
            ['sub_3', 'sub_5']
        )
        self.assertListEqual(
            action_per_sub[0].y.tolist(),
            [19, 19]
        )
        self.assertListEqual(
            action_per_line[0].x.tolist(),
            ['3_6_15', '9_10_12']
        )
        self.assertListEqual(
            action_per_line[0].y.tolist(),
            [13, 12]
        )

        self.assertListEqual(
            self.episode_analytics.action_data_table.action_id[:5].tolist(),
            [0, 1, 1, 2, 3]
        )
        self.assertListEqual(
            self.episode_analytics.action_data_table.distance[:5].tolist(),
            [1, 2, 2, 0, 3]
        )






"""
    This files handles the generic information about the agent of reference of the selected scenario
    and let choose and compute study agent information.
"""
import datetime as dt
from pathlib import Path

import plotly.graph_objects as go
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

from grid2viz.src.kpi import EpisodeTrace
from grid2viz.src.kpi import actions_model
from grid2viz.src.manager import (
    make_episode,
    make_network_agent_overview,
    grid2viz_home_directory,
)
from grid2viz.src.utils.callbacks_helpers import toggle_modal_helper
from grid2viz.src.utils.common_graph import make_action_ts, make_rewards_ts
from grid2viz.src.utils.constants import DONT_SHOW_FILENAME
from grid2viz.src.utils.graph_utils import (
    get_axis_relayout,
    relayout_callback,
    max_or_zero,
)


def register_callbacks_macro(app):

    # Synchronized graphs
    @app.callback(
        [
            Output("rewards_timeserie", "figure"),
            Output("cumulated_rewards_timeserie", "figure"),
            Output("overflow_graph_study", "figure"),
            Output("usage_rate_graph_study", "figure"),
            Output("action_timeserie", "figure"),
        ],
        [
            Input("agent_study", "modified_timestamp"),
            Input("agent_ref", "data"),
            Input("relayoutStoreMacro", "data"),
        ],
        [
            State("rewards_timeserie", "figure"),
            State("cumulated_rewards_timeserie", "figure"),
            State("overflow_graph_study", "figure"),
            State("usage_rate_graph_study", "figure"),
            State("action_timeserie", "figure"),
            State("scenario", "data"),
            State("agent_study", "data"),
            State("relayoutStoreMacro", "modified_timestamp"),
        ],
    )
    def update_synchronized_figures(
        agent_study_ts,
        ref_agent,
        relayout_data_store,
        rew_figure,
        cumrew_figure,
        overflow_figure,
        usage_rate_figure,
        actions_figure,
        scenario,
        study_agent,
        relayoutStoreMacro_ts,
    ):

        figures = [
            rew_figure,
            cumrew_figure,
            overflow_figure,
            usage_rate_figure,
            actions_figure,
        ]

        episode = make_episode(study_agent, scenario)

        if agent_study_ts is not None and relayoutStoreMacro_ts is not None:
            condition = (
                relayout_data_store is not None
                and relayout_data_store["relayout_data"]
                and relayoutStoreMacro_ts > agent_study_ts
            )
        else:
            condition = (
                relayout_data_store is not None and relayout_data_store["relayout_data"]
            )
        if condition:
            relayout_data = relayout_data_store["relayout_data"]
            relayouted = False
            for figure in figures:
                axis_layout = get_axis_relayout(figure, relayout_data)
                if axis_layout is not None:
                    figure["layout"].update(axis_layout)
                    relayouted = True
            if relayouted:
                return figures
        new_reward_fig, new_cumreward_fig = make_rewards_ts(
            study_agent, ref_agent, scenario, rew_figure, cumrew_figure
        )

        overflow_figure["data"] = episode.total_overflow_trace.copy()
        for event in ["maintenance", "hazard", "attacks"]:
            func = getattr(EpisodeTrace, f"get_{event}_trace")
            traces = func(episode, ["total"])
            if len(traces) > 0:
                traces[0].update({"name": event.capitalize()})
                overflow_figure["data"].append(traces[0])

        usage_rate_figure["data"] = episode.usage_rate_trace

        new_action_fig = make_action_ts(
            study_agent, ref_agent, scenario, actions_figure["layout"]
        )

        return (
            new_reward_fig,
            new_cumreward_fig,
            overflow_figure,
            usage_rate_figure,
            new_action_fig,
        )

    @app.callback(
        Output("agent_study_pie_chart", "figure"),
        [Input("agent_study", "data")],
        [State("agent_study_pie_chart", "figure"), State("scenario", "data")],
    )
    def update_action_repartition_pie(study_agent, figure, scenario):
        new_episode = make_episode(study_agent, scenario)
        figure["data"] = action_repartition_pie(new_episode)
        figure["layout"].update(
            actions_model.update_layout(
                figure["data"][0].values == (0, 0, 0), "No Actions for this Agent"
            )
        )
        return figure

    def action_repartition_pie(agent):
        nb_actions = agent.action_data_table[
            ["action_line", "action_subs", "action_redisp"]
        ].sum()
        return [
            go.Pie(
                labels=[
                    "Actions on Lines",
                    "Actions on Substations",
                    "Redispatching Actions",
                ],
                values=[
                    nb_actions["action_line"],
                    nb_actions["action_subs"],
                    nb_actions["action_redisp"],
                ],
            )
        ]

    # @app.callback(
    #     Output("network_actions", "figure"),
    #     [Input("agent_study", "data")],
    #     [State("scenario", "data")]
    # )
    # def update_network_graph(study_agent, scenario):
    #     episode = make_episode(study_agent, scenario)
    #
    #     return make_network_agent_overview(episode)

    @app.callback(
        Output("timeseries_table", "data"),
        [Input("rewards_timeserie", "clickData"), Input("select_study_agent", "value")],
        [State("timeseries_table", "data"), State("agent_study", "data")],
    )
    def add_timestamp(click_data, new_agent, data, agent_stored):
        if new_agent != agent_stored:
            return []
        if click_data is None:
            if data is not None:
                return data
            else:
                return []
        time_stamp_str = click_data["points"][0]["x"]
        try:
            dt.datetime.strptime(time_stamp_str, "%Y-%m-%d %H:%M")
        except ValueError:
            time_stamp_str = time_stamp_str + " 00:00"
        new_data = {"Timestamps": time_stamp_str}
        if new_data not in data:
            data.append(new_data)
        return data

    @app.callback(
        Output("user_timestamps_store", "data"), [Input("timeseries_table", "data")]
    )
    def update_user_timestamps_store(timestamps):
        if timestamps is None:
            raise PreventUpdate
        return [
            dict(label=timestamp["Timestamps"], value=timestamp["Timestamps"])
            for timestamp in timestamps
        ]

    @app.callback(
        Output("relayoutStoreMacro", "data"),
        [
            Input("usage_rate_graph_study", "relayoutData"),
            Input("action_timeserie", "relayoutData"),
            Input("overflow_graph_study", "relayoutData"),
            Input("rewards_timeserie", "relayoutData"),
            Input("cumulated_rewards_timeserie", "relayoutData"),
        ],
        [State("relayoutStoreMacro", "data")],
    )
    def relayout_store(*args):
        return relayout_callback(*args)

    @app.callback(
        [
            Output("indicator_score_output", "children"),
            Output("indicator_nb_overflow", "children"),
            Output("indicator_nb_action", "children"),
        ],
        [Input("agent_study", "data"), Input("scenario", "data")],
    )
    def update_nbs(study_agent, scenario):
        new_episode = make_episode(study_agent, scenario)
        score = f"{get_score_agent(new_episode):,}"
        nb_overflow = f"{get_nb_overflow_agent(new_episode):,}"
        nb_action = f"{get_nb_action_agent(new_episode):,}"

        return score, nb_overflow, nb_action

    def get_score_agent(agent):
        score = agent.meta["cumulative_reward"]
        return round(score)

    def get_nb_overflow_agent(agent):
        return agent.total_overflow_ts["value"].sum()

    def get_nb_action_agent(agent):
        return int(
            agent.action_data_table[["action_line", "action_subs"]].sum(axis=1).sum()
        )

    # @app.callback(
    #     Output("agent_study", "data"),
    #     [Input('agent_log_selector', 'value')],
    #     [State("agent_study", "data"),
    #      State("scenario", "data")],
    # )
    # def update_study_agent(study_agent, stored_agent, scenario):
    #     if study_agent == stored_agent:
    #         raise PreventUpdate
    #     make_episode(study_agent, scenario)
    #     return study_agent

    action_table_name_converter = dict(
        timestep="Timestep",
        timestamp="Timestamp",
        timestep_reward="Reward",
        action_line="Action on line",
        action_subs="Action on sub",
        action_redisp="Action of redispatch",
        redisp_impact="Redispatch impact",
        line_name="Line name",
        sub_name="Sub name",
        gen_name="Gen name",
        action_id="Action id",
        distance="Topological distance",
    )

    @app.callback(
        [Output("inspector_datable", "columns"), Output("inspector_datable", "data")],
        [Input("agent_study", "data"), Input("scenario", "data")],
    )
    def update_agent_log_action_table(study_agent, scenario):
        new_episode = make_episode(study_agent, scenario)
        table = actions_model.get_action_table_data(new_episode)
        table["id"] = table["timestep"]
        table.set_index("id", inplace=True, drop=False)
        cols_to_exclude = ["id", "lines_modified", "subs_modified", "gens_modified"]
        cols = [
            {"name": action_table_name_converter[col], "id": col}
            for col in table.columns
            if col not in cols_to_exclude
        ]
        return cols, table.to_dict("record")

    @app.callback(
        [
            Output("distribution_substation_action_chart", "figure"),
            Output("distribution_line_action_chart", "figure"),
            Output("distribution_redisp_action_chart", "figure"),
        ],
        [Input("agent_study", "data"), Input("agent_ref", "data")],
        [
            State("distribution_substation_action_chart", "figure"),
            State("distribution_line_action_chart", "figure"),
            State("distribution_redisp_action_chart", "figure"),
            State("scenario", "data"),
        ],
    )
    def update_agent_log_action_graphs(
        study_agent, ref_agent, figure_sub, figure_switch_line, figure_redisp, scenario
    ):
        new_episode = make_episode(study_agent, scenario)
        ref_episode = make_episode(ref_agent, scenario)
        y_max = None
        figure_sub["data"] = actions_model.get_action_per_sub(new_episode)
        if len(figure_sub["data"][0]["x"]) != 0:
            figure_sub["data"].append(actions_model.get_action_per_sub(ref_episode)[0])
            y_max = max(map(max_or_zero, [trace.y for trace in figure_sub["data"]])) + 1
        figure_switch_line["data"] = actions_model.get_action_per_line(new_episode)
        if len(figure_switch_line["data"][0]["x"]) != 0:
            figure_switch_line["data"].append(
                actions_model.get_action_per_line(ref_episode)[0]
            )
            if y_max is None:
                y_max = (
                    max(
                        map(
                            max_or_zero,
                            [trace.y for trace in figure_switch_line["data"]],
                        )
                    )
                    + 1
                )
            if (
                max(map(max_or_zero, [trace.y for trace in figure_switch_line["data"]]))
                > y_max
            ):
                y_max = (
                    max(
                        map(
                            max_or_zero,
                            [trace.y for trace in figure_switch_line["data"]],
                        )
                    )
                    + 1
                )
        figure_redisp["data"] = actions_model.get_action_redispatch(new_episode)
        if len(figure_redisp["data"][0]["x"]) != 0:
            figure_redisp["data"].append(
                actions_model.get_action_redispatch(ref_episode)[0]
            )
            if y_max is None:
                y_max = (
                    max(map(max_or_zero, [trace.y for trace in figure_redisp["data"]]))
                    + 1
                )
            if (
                max(map(max_or_zero, [trace.y for trace in figure_redisp["data"]]))
                > y_max
            ):
                y_max = (
                    max(map(max_or_zero, [trace.y for trace in figure_redisp["data"]]))
                    + 1
                )

        figure_sub["layout"].update(
            actions_model.update_layout(
                len(figure_sub["data"][0]["x"]) == 0,
                "No Actions on subs for this Agent",
            )
        )
        figure_switch_line["layout"].update(
            actions_model.update_layout(
                len(figure_switch_line["data"][0]["x"]) == 0,
                "No Actions on lines for this Agent",
            )
        )
        figure_redisp["layout"].update(
            actions_model.update_layout(
                len(figure_redisp["data"][0]["x"]) == 0,
                "No redispatching actions for this Agent",
            )
        )

        if y_max:
            figure_sub["layout"]["yaxis"].update(range=[0, y_max])
            figure_switch_line["layout"]["yaxis"].update(range=[0, y_max])
            figure_redisp["layout"]["yaxis"].update(range=[0, y_max])

        return figure_sub, figure_switch_line, figure_redisp

    @app.callback(
        Output("tooltip_table", "children"),
        [Input("inspector_datable", "active_cell")],
        [
            State("agent_study", "data"),
            State("scenario", "data"),
            State("inspector_datable", "data"),
        ],
    )
    def update_more_info(active_cell, study_agent, scenario, data):
        if active_cell is None:
            raise PreventUpdate
        new_episode = make_episode(study_agent, scenario)
        row_id = active_cell["row_id"]
        act = new_episode.actions[row_id]
        return str(act)

    @app.callback(
        [
            Output("modal_macro", "is_open"),
            Output("dont_show_again_div_macro", "className"),
        ],
        [Input("close_macro", "n_clicks"), Input("page_help", "n_clicks")],
        [State("modal_macro", "is_open"), State("dont_show_again_macro", "checked")],
    )
    def toggle_modal(close_n_clicks, open_n_clicks, is_open, dont_show_again):
        dsa_filepath = Path(grid2viz_home_directory) / DONT_SHOW_FILENAME("macro")
        return toggle_modal_helper(
            close_n_clicks,
            open_n_clicks,
            is_open,
            dont_show_again,
            dsa_filepath,
            "page_help",
        )

    @app.callback(Output("modal_image_macro", "src"), [Input("url", "pathname")])
    def show_image(pathname):
        return app.get_asset_url("screenshots/agent_overview.png")



import dash_antd_components as dac
import dash_bootstrap_components as dbc
import dash_core_components as dcc
import dash_html_components as html

from grid2viz.src.manager import make_network, make_episode


def choose_assist_line(episode, network_graph):
    return html.Div(
        id="choose_assist_line",
        className="lineBlock card",
        children=[
            html.H4("Choose or Assist"),
            html.Div(
                className="card-body row",
                children=[
                    html.Div(
                        className="col-7",
                        children=[
                            html.H5("Network at time step t"),
                            dcc.Graph(
                                id="network_graph_choose",
                                figure=network_graph,
                            ),
                        ],
                    ),
                    html.Div(
                        className="col-5",
                        children=[
                            dbc.Tabs(
                                children=[
                                    dbc.Tab(
                                        label="Choose",
                                        labelClassName="fas fa-user",
                                        children=[
                                            dbc.Tabs(
                                                children=[
                                                    dbc.Tab(
                                                        label="Dropdowns",
                                                        children=[
                                                            dbc.Tabs(
                                                                children=[
                                                                    dbc.Tab(
                                                                        label="Lines",
                                                                        children=[
                                                                            html.P(
                                                                                "Choose a line to act on:",
                                                                                className="mt-1",
                                                                            ),
                                                                            dac.Select(
                                                                                id="select_lines_simulation",
                                                                                options=[
                                                                                    {
                                                                                        "label": line_name,
                                                                                        "value": line_name,
                                                                                    }
                                                                                    for line_name in episode.line_names
                                                                                ],
                                                                                mode="default",
                                                                                value=episode.line_names[
                                                                                    0
                                                                                ],
                                                                            ),
                                                                            html.P(
                                                                                "Choose an action:",
                                                                                className="mt-1",
                                                                            ),
                                                                            dac.Radio(
                                                                                options=[
                                                                                    {
                                                                                        "label": "Set",
                                                                                        "value": "Set",
                                                                                    },
                                                                                    {
                                                                                        "label": "Change",
                                                                                        "value": "Change",
                                                                                    },
                                                                                ],
                                                                                value="Set",
                                                                                id="select_action_lines",
                                                                                buttonStyle="solid",
                                                                            ),
                                                                        ],
                                                                    ),
                                                                    dbc.Tab(
                                                                        label="Subs",
                                                                        children=[
                                                                            html.P(
                                                                                "Choose a substation to act on:"
                                                                            ),
                                                                            dac.Select(
                                                                                id="select_subs_simulation",
                                                                                options=[
                                                                                    {
                                                                                        "label": name,
                                                                                        "value": name,
                                                                                    }
                                                                                    for name in episode.name_sub
                                                                                ],
                                                                                mode="default",
                                                                                value=episode.name_sub[
                                                                                    0
                                                                                ],
                                                                            ),
                                                                            html.P(
                                                                                "Choose an action:"
                                                                            ),
                                                                            dac.Radio(
                                                                                options=[
                                                                                    {
                                                                                        "label": "Set",
                                                                                        "value": "Set",
                                                                                    },
                                                                                    {
                                                                                        "label": "Change",
                                                                                        "value": "Change",
                                                                                    },
                                                                                ],
                                                                                value="Set",
                                                                                id="select_action_subs",
                                                                                buttonStyle="solid",
                                                                            ),
                                                                        ],
                                                                    ),
                                                                    dbc.Tab(
                                                                        label="Gens",
                                                                        children=[
                                                                            html.P(
                                                                                "Choose a generator to act on:"
                                                                            ),
                                                                            dac.Select(
                                                                                id="select_gens_simulation",
                                                                                options=[
                                                                                    {
                                                                                        "label": prod_name,
                                                                                        "value": prod_name,
                                                                                    }
                                                                                    for prod_name in episode.prod_names
                                                                                ],
                                                                                mode="default",
                                                                                value=episode.prod_names[
                                                                                    0
                                                                                ],
                                                                            ),
                                                                            html.P(
                                                                                "Choose an action:"
                                                                            ),
                                                                            dac.Radio(
                                                                                options=[
                                                                                    {
                                                                                        "label": "Redispatch",
                                                                                        "value": "Redispatch",
                                                                                    },
                                                                                ],
                                                                                value="Redispatch",
                                                                                id="select_action_gens",
                                                                                buttonStyle="solid",
                                                                            ),
                                                                        ],
                                                                    ),
                                                                ]
                                                            )
                                                        ],
                                                    ),
                                                    dbc.Tab(
                                                        label="Dict",
                                                        children=[
                                                            html.P(
                                                                "Enter the action dictionary:"
                                                            ),
                                                            dbc.Textarea(
                                                                className="mb-3",
                                                                placeholder="{set_line_status: []}",
                                                            ),
                                                        ],
                                                    ),
                                                ]
                                            ),
                                        ],
                                    ),
                                    dbc.Tab(
                                        label="Assist",
                                        labelClassName="fas fa-robot",
                                        children=["content"],
                                    ),
                                ]
                            ),
                            dbc.Button(
                                "Simulate",
                                id="simulate_action",
                                color="danger",
                                className="mt-3 mb-3",
                            ),
                            html.P("Action dictionary:"),
                            html.Div(id="action_dictionary"),
                        ],
                    ),
                ],
            ),
        ],
    )


def compare_line(network_graph):
    return html.Div(
        id="compare_line",
        className="lineBlock card",
        children=[
            html.H4("Compare"),
            html.Div(
                className="card-body row",
                children=[
                    html.Div(
                        className="col-9",
                        children=[
                            dac.Radio(
                                options=[
                                    {"label": "New State t+1", "value": "new"},
                                    {"label": "Old State t+1", "value": "old"},
                                    {"label": "Delta", "value": "compare"},
                                ],
                                value="new",
                                id="compare_states_radio",
                                buttonStyle="solid",
                            ),
                            dcc.Graph(id="network_graph_compare", figure=network_graph),
                        ],
                    ),
                    html.Div(
                        className="col-3",
                        children=[
                            html.H5("Agent's KPIs"),
                            html.Div(
                                className="mb-4",
                                children=[
                                    html.P(
                                        id="agent_reward",
                                        className="border-bottom h3 mb-0 text-right",
                                        children="0",
                                    ),
                                    html.P(
                                        className="text-muted",
                                        children="Agent's reward",
                                    ),
                                ],
                            ),
                            html.Div(
                                className="mb-4",
                                children=[
                                    html.P(
                                        id="agent_rho",
                                        className="border-bottom h3 mb-0 text-right",
                                        children="0",
                                    ),
                                    html.P(
                                        className="text-muted",
                                        children="Agent's max rho",
                                    ),
                                ],
                            ),
                            html.Div(
                                className="mb-4",
                                children=[
                                    html.P(
                                        id="agent_overflows",
                                        className="border-bottom h3 mb-0 text-right",
                                        children="0",
                                    ),
                                    html.P(
                                        className="text-muted",
                                        children="Agent's overflows",
                                    ),
                                ],
                            ),
                            html.Div(
                                className="mb-4",
                                children=[
                                    html.P(
                                        id="agent_losses",
                                        className="border-bottom h3 mb-0 text-right",
                                        children="0",
                                    ),
                                    html.P(
                                        className="text-muted",
                                        children="Agent's losses",
                                    ),
                                ],
                            ),
                            html.H5("New Action's KPIs"),
                            html.Div(
                                className="mb-4",
                                children=[
                                    html.P(
                                        id="new_action_reward",
                                        className="border-bottom h3 mb-0 text-right",
                                        children="0",
                                    ),
                                    html.P(
                                        className="text-muted",
                                        children="New Action reward",
                                    ),
                                ],
                            ),
                            html.Div(
                                className="mb-4",
                                children=[
                                    html.P(
                                        id="new_action_rho",
                                        className="border-bottom h3 mb-0 text-right",
                                        children="0",
                                    ),
                                    html.P(
                                        className="text-muted",
                                        children="New Action max rho",
                                    ),
                                ],
                            ),
                            html.Div(
                                className="mb-4",
                                children=[
                                    html.P(
                                        id="new_action_overflows",
                                        className="border-bottom h3 mb-0 text-right",
                                        children="0",
                                    ),
                                    html.P(
                                        className="text-muted",
                                        children="New Action overflows",
                                    ),
                                ],
                            ),
                            html.Div(
                                className="mb-4",
                                children=[
                                    html.P(
                                        id="new_action_losses",
                                        className="border-bottom h3 mb-0 text-right",
                                        children="0",
                                    ),
                                    html.P(
                                        className="text-muted",
                                        children="New Action losses",
                                    ),
                                ],
                            ),
                        ],
                    ),
                ],
            ),
        ],
    )


def layout(scenario, studied_agent):
    episode = make_episode(studied_agent, scenario)
    # TODO : center en correct time step
    network_graph = make_network(episode).plot_obs(observation=episode.observations[0])
    return html.Div(
        id="simulation_page",
        children=[
            choose_assist_line(episode, network_graph),
            compare_line(network_graph),
        ],
    )



import argparse
import os

## A bug can appear with MacOSX if matplotlib is not set to a non-interactive mode
# issue: https://github.com/matplotlib/matplotlib/issues/14304/
import matplotlib

matplotlib.use("agg")

CONFIG_FILE_CONTENT = """
# This file have been automatically generated, please do not modify it. 
# You can remove it once done with the application.
[DEFAULT]
agents_dir={agents_dir}
env_dir={env_dir}
n_cores={n_cores}
# This file will be re generated to each call of "python -m grid2viz.main"
"""

ARG_AGENTS_PATH_DESC = (
    "The path where the log of the Agents experiences are stored."
    " (default to None to study the example agents provided with the package)"
)
ARG_ENV_PATH_DESC = (
    "The path where the environment config is stored."
    " (default to None to use the provided default environment)"
)
ARG_PORT_DESC = "The port to serve grid2viz on." " (default to 8050)"
ARG_DEBUG_DESC = "Enable debug mode for developers." " (default to False)"
ARG_N_CORES_DESC = "The number of cores to use for the first loading of the best agents of each scenario"

ARG_CACHE_DESC = "True if you want to build all the cache data for all agents at once before relaunching grid2viz"


def main():
    parser_main = argparse.ArgumentParser(description="Grid2Viz")
    parser_main.add_argument(
        "--agents_path",
        default=None,
        required=False,
        type=str,
        help=ARG_AGENTS_PATH_DESC,
    )
    parser_main.add_argument(
        "--env_path", default=None, required=False, type=str, help=ARG_ENV_PATH_DESC
    )
    parser_main.add_argument(
        "--port", default=8050, required=False, type=int, help=ARG_PORT_DESC
    )
    parser_main.add_argument("--debug", action="store_true", help=ARG_DEBUG_DESC)

    parser_main.add_argument("--n_cores", default=2, type=int, help=ARG_N_CORES_DESC)
    parser_main.add_argument("--cache", default=False, type=bool, help=ARG_CACHE_DESC)

    args = parser_main.parse_args()

    pkg_root_dir = os.path.dirname(os.path.abspath(__file__))
    os.environ["GRID2VIZ_ROOT"] = pkg_root_dir
    config_path = os.path.join(pkg_root_dir, "config.ini")

    if args.agents_path is not None:
        agents_dir = os.path.abspath(args.agents_path)
    else:
        agents_dir = os.path.join(pkg_root_dir, "data", "agents")
        print("Using default agents logs at {}".format(agents_dir))

    if args.env_path is not None:
        env_dir = os.path.abspath(args.env_path)
    else:
        env_dir = os.path.join(pkg_root_dir, "data", "env_conf")
        print("Using default environment at {}".format(env_dir))

    n_cores = args.n_cores

    with open(config_path, "w") as f:
        f.write(
            CONFIG_FILE_CONTENT.format(
                agents_dir=agents_dir, env_dir=env_dir, n_cores=n_cores
            )
        )

    is_makeCache_only = args.cache

    # Inline import to load app only now
    if is_makeCache_only:
        make_cache()
    else:
        from grid2viz.app import app_run

        app_run(args.port, args.debug)


def make_cache():
    from grid2viz.src.manager import (
        scenarios,
        agents,
        make_episode_without_decorate,
        n_cores,
        retrieve_episode_from_disk,
        save_in_ram_cache,
        cache_dir,
    )

    from pathos.multiprocessing import ProcessPool

    if not os.path.exists(cache_dir):
        print("Starting Multiprocessing for reading the best agent of each scenario")

    # TODO: tous les agents n'ont pas forcément tourner sur exactement tous les mêmes scenarios
    # Eviter une erreur si un agent n'a pas tourné sur un scenario
    agent_scenario_list = [
        (agent, scenario) for agent in agents for scenario in scenarios
    ]

    agents_data = []
    if n_cores == 1:  # no multiprocess useful for debug if needed
        i = 0
        for agent_scenario in agent_scenario_list:
            agents_data.append(
                make_episode_without_decorate(agent_scenario[0], agent_scenario[1])
            )
            i += 1
    else:
        pool = ProcessPool(n_cores)
        agents_data = list(
            pool.imap(
                make_episode_without_decorate,
                [agent_scenario[0] for agent_scenario in agent_scenario_list],  # agents
                [agent_scenario[1] for agent_scenario in agent_scenario_list],
            )
        )  # scenarios #we go over all agents and all scenarios for each agent
        pool.close()
        print("Multiprocessing done")

    #####
    # saving data on disk
    i = 0
    for agent_scenario in agent_scenario_list:
        print(i)
        agent = agent_scenario[0]
        episode_name = agent_scenario[1]
        agent_episode = agents_data[i]
        if agent_episode is not None:
            episode_data = retrieve_episode_from_disk(
                agent_episode.episode_name, agent_episode.agent
            )

            agent_episode.decorate(episode_data)
            save_in_ram_cache(
                agent_episode.episode_name, agent_episode.agent, agent_episode
            )
        i += 1


if __name__ == "__main__":
    main()
